"""Test cases for Zurich Instruments HDAWG."""
import json
import logging
import unittest
from unittest.mock import call, Mock, ANY
import warnings

import numpy as np

import qmi
from qmi.instruments.zurich_instruments import ZurichInstruments_Hdawg
import qmi.instruments.zurich_instruments.hdawg as hdawg
from qmi.core.exceptions import QMI_InvalidOperationException

_DEVICE_HOST = "localhost"
_DEVICE_PORT = 12345
_DEVICE_NAME = "DEV8888"


class TestHDAWGInit(unittest.TestCase):
    """Testcase for HDAWG initialization."""

    def setUp(self):
        logging.getLogger("qmi.core.instrument").setLevel(logging.CRITICAL)

        qmi.start("test_hdawg_openclose", init_logging=False)
        self._awg_module = Mock()
        self._awg_module.finished = Mock()

        self._daq_server = Mock()
        self._daq_server.awgModule = Mock(return_value=self._awg_module)

        hdawg.zhinst = Mock()
        hdawg.zhinst.ziPython = Mock()
        hdawg.zhinst.utils = Mock()
        hdawg.zhinst.ziPython.ziDAQServer = Mock(return_value=self._daq_server)

        self._hdawg = qmi.make_instrument(
            "HDAWG",
            ZurichInstruments_Hdawg,
            server_host=_DEVICE_HOST,
            server_port=_DEVICE_PORT,
            device_name=_DEVICE_NAME
        )

    def tearDown(self):
        with warnings.catch_warnings():
            # Suppress warnings when instrument not correctly closed.
            warnings.simplefilter("ignore", ResourceWarning)
            qmi.stop()
        logging.getLogger("qmi.core.instrument").setLevel(logging.NOTSET)

    def test_openclose(self):
        """Nominal open/close sequence."""
        self._awg_module.finished.side_effect = [True, False, False, True]

        self._hdawg.open()
        self._hdawg.close()

        expected_daq_server_calls = [
            # open()
            call.connectDevice(_DEVICE_NAME, "1GbE"),
            call.awgModule(),
            # close()
            call.disconnect()
        ]
        self.assertEqual(self._daq_server.mock_calls, expected_daq_server_calls)

        expected_awg_module_calls = [
            # open()
            call.set("awgModule/device", _DEVICE_NAME),
            call.set("awgModule/index", 0),
            call.finished(),
            call.execute(),
            call.finished(),
            # close()
            call.finished(),
            call.finish(),
            call.finished()
        ]
        self.assertEqual(self._awg_module.mock_calls, expected_awg_module_calls)

    def test_failed_open1(self):
        """Failed open sequence."""
        self._awg_module.finished.side_effect = [False]

        with self.assertRaises(AssertionError):
            self._hdawg.open()

        expected_daq_server_calls = [
            call.connectDevice(_DEVICE_NAME, "1GbE"),
            call.awgModule()
        ]
        self.assertEqual(self._daq_server.mock_calls, expected_daq_server_calls)

        expected_awg_module_calls = [
            call.set("awgModule/device", _DEVICE_NAME),
            call.set("awgModule/index", 0),
            call.finished()
        ]
        self.assertEqual(self._awg_module.mock_calls, expected_awg_module_calls)

    def test_failed_open2(self):
        """Failed open sequence."""
        self._awg_module.finished.side_effect = [True, True]

        with self.assertRaises(AssertionError):
            self._hdawg.open()

        expected_daq_server_calls = [
            call.connectDevice(_DEVICE_NAME, "1GbE"),
            call.awgModule()
        ]
        self.assertEqual(self._daq_server.mock_calls, expected_daq_server_calls)

        expected_awg_module_calls = [
            call.set("awgModule/device", _DEVICE_NAME),
            call.set("awgModule/index", 0),
            call.finished(),
            call.execute(),
            call.finished()
        ]
        self.assertEqual(self._awg_module.mock_calls, expected_awg_module_calls)

    def test_failed_close1(self):
        """Failed close sequence."""
        self._awg_module.finished.side_effect = [True, False, True]

        self._hdawg.open()
        with self.assertRaises(AssertionError):
            self._hdawg.close()

        expected_daq_server_calls = [
            # open()
            call.connectDevice(_DEVICE_NAME, "1GbE"),
            call.awgModule()
        ]
        self.assertEqual(self._daq_server.mock_calls, expected_daq_server_calls)

        expected_awg_module_calls = [
            # open()
            call.set("awgModule/device", _DEVICE_NAME),
            call.set("awgModule/index", 0),
            call.finished(),
            call.execute(),
            call.finished(),
            # close()
            call.finished()
        ]
        self.assertEqual(self._awg_module.mock_calls, expected_awg_module_calls)

    def test_failed_close2(self):
        """Failed close sequence."""
        self._awg_module.finished.side_effect = [True, False, False, False]

        self._hdawg.open()
        with self.assertRaises(AssertionError):
            self._hdawg.close()

        expected_daq_server_calls = [
            # open()
            call.connectDevice(_DEVICE_NAME, "1GbE"),
            call.awgModule()
        ]
        self.assertEqual(self._daq_server.mock_calls, expected_daq_server_calls)

        expected_awg_module_calls = [
            # open()
            call.set("awgModule/device", _DEVICE_NAME),
            call.set("awgModule/index", 0),
            call.finished(),
            call.execute(),
            call.finished(),
            # close()
            call.finished(),
            call.finish(),
            call.finished()
        ]
        self.assertEqual(self._awg_module.mock_calls, expected_awg_module_calls)

    def test_failed_open_notclosed(self):
        """Open device that is already open."""
        self._awg_module.finished.side_effect = [True, False, False, True]
        self._hdawg.open()

        with self.assertRaises(QMI_InvalidOperationException):
            self._hdawg.open()
        self._hdawg.close()

    def test_failed_close_notopen(self):
        """Close device that is not open."""
        with self.assertRaises(QMI_InvalidOperationException):
            self._hdawg.close()


class TestHDAWG(unittest.TestCase):
    """Testcase for HDAWG functions."""

    def setUp(self):
        logging.getLogger("qmi.instruments.zurich_instruments.hdawg").setLevel(logging.CRITICAL)

        qmi.start("TestHDAWGMethods", init_logging=False)
        self._awg_module = Mock()
        self._awg_module.finished = Mock()
        self._awg_module.finished.side_effect = [True, False, False, True]
        self._awg_module.getInt = Mock()
        self._awg_module.getDouble = Mock()

        self._daq_server = Mock()
        self._daq_server.awgModule = Mock(return_value=self._awg_module)

        hdawg.zhinst = Mock()
        hdawg.zhinst.ziPython = Mock()
        hdawg.zhinst.utils = Mock()
        hdawg.zhinst.ziPython.ziDAQServer = Mock(return_value=self._daq_server)
        hdawg._COMPILE_TIMEOUT = 0.0
        hdawg._UPLOAD_TIMEOUT = 0.0

        self._hdawg = qmi.make_instrument(
            "HDAWG",
            ZurichInstruments_Hdawg,
            server_host=_DEVICE_HOST,
            server_port=_DEVICE_PORT,
            device_name=_DEVICE_NAME
        )
        self._hdawg.open()
        self._awg_module.reset_mock()
        self._daq_server.reset_mock()

    def _check_get_value_int(self, node_path):
        expected_calls = [
            call.getInt("/{}/{}".format(_DEVICE_NAME, node_path))
        ]
        self._daq_server.assert_has_calls(expected_calls)

    def _check_get_value_float(self, node_path):
        expected_calls = [
            call.getDouble("/{}/{}".format(_DEVICE_NAME, node_path))
        ]
        self._daq_server.assert_has_calls(expected_calls)

    def _check_set_value(self, node_path, value):
        expected_calls = [
            call.set("/{}/{}".format(_DEVICE_NAME, node_path), value)
        ]
        self._daq_server.assert_has_calls(expected_calls)

    def _check_set_value_int(self, node_path, value):
        expected_calls = [
            call.setInt("/{}/{}".format(_DEVICE_NAME, node_path), value)
        ]
        self._daq_server.assert_has_calls(expected_calls)

    def _check_get_value_double(self, node_path):
        expected_calls = [
            call.getDouble("/{}/{}".format(_DEVICE_NAME, node_path))
        ]
        self._daq_server.assert_has_calls(expected_calls)

    def _check_set_value_double(self, node_path, value):
        expected_calls = [
            call.setDouble("/{}/{}".format(_DEVICE_NAME, node_path), value)
        ]
        self._daq_server.assert_has_calls(expected_calls)

    def tearDown(self):
        self._hdawg.close()
        qmi.stop()
        logging.getLogger("qmi.instruments.zurich_instruments.hdawg").setLevel(logging.NOTSET)

    def test_generic_set(self):
        """Test setter."""
        node = "my/test/node"
        value = "value"
        self._hdawg.set_node_value(node, value)
        self._check_set_value(node, value)

    def test_generic_get_int(self):
        """Test getter."""
        node = "my/test/node"
        value = 123
        self._daq_server.getInt.return_value = value
        result = self._hdawg.get_node_int(node)
        self._check_get_value_int(node)
        self.assertEqual(result, value)

    def test_generic_set_int(self):
        """Test setter."""
        node = "my/test/node"
        value = 123
        self._hdawg.set_node_int(node, value)
        self._check_set_value_int(node, value)

    def test_generic_get_double(self):
        """Test getter."""
        node = "my/test/node"
        value = 123.456
        self._daq_server.getDouble.return_value = value
        result = self._hdawg.get_node_double(node)
        self._check_get_value_double(node)
        self.assertEqual(result, value)

    def test_generic_set_double(self):
        """Test setter."""
        node = "my/test/node"
        value = 123.345
        self._hdawg.set_node_double(node, value)
        self._check_set_value_double(node, value)

    def test_compile_and_upload(self):
        """Test compile sequence."""
        source = """
        wave w = "$FILE";
        var n = 0.0;
        for(i = 0; i < $COUNT; i++) {
            n += $INCR;
        }
        """

        replacements = {
            "$FILE": "w_a",
            "$COUNT": 100,
            "$INCR": 3.14
        }

        expected_source = """
        wave w = "w_a";
        var n = 0.0;
        for(i = 0; i < 100; i++) {
            n += 3.14;
        }
        """

        self._awg_module.getInt.side_effect = [0, 0]  # successful compilation; upload completed
        self._awg_module.getDouble.side_effect = [1.0]  # upload progress

        expected_calls = [
            call.set("awgModule/compiler/sourcestring", expected_source),
            call.getInt("awgModule/compiler/status"),
            call.getDouble("awgModule/progress"),
            call.getInt("awgModule/elf/status"),
        ]

        self._hdawg.compile_and_upload(source, replacements)
        self.assertEqual(self._awg_module.mock_calls, expected_calls)
        self.assertTrue(self._hdawg.compilation_successful())

    def test_compile_empty_fails(self):
        """Test that empty programs are not accepted."""
        with self.assertRaises(ValueError):
            self._hdawg.compile_and_upload("")

        with self.assertRaises(ValueError):
            self._hdawg.compile_and_upload("// Hello")

        with self.assertRaises(ValueError):
            self._hdawg.compile_and_upload("/* Hello */")

        with self.assertRaises(ValueError):
            self._hdawg.compile_and_upload("""
                // Hello

                /* Hello */
                """)

    def test_valid_replacements(self):
        """Test replacement patterns"""
        cases = [
            ("par = $PAR;", {"$PAR": "value"}, "par = value;"),  # standard replace
            ("PAR = $PAR;", {"$PAR": "value"}, "PAR = value;"),  # $ distinguishes replacement from sequencer code
            ("par = $A;", {"$A": 1}, "par = 1;"),  # upper case parameter name
            ("par = $a;", {"$a": 3.14}, "par = 3.14;"),  # lower case parameter name
            ("par = $PAR1;", {"$PAR1": "value"}, "par = value;"),  # numbers in parameter name
            ("par = $A1;", {"$A1": 2}, "par = 2;"),
            ("par = $a1;", {"$a1": 3.14}, "par = 3.14;"),
            ("par = $PAR_;", {"$PAR_": "value"}, "par = value;"),  # underscore in parameter name
            ("par = $A_;", {"$A_": 3}, "par = 3;"),
            ("par = $a_;", {"$a_": 3.14}, "par = 3.14;"),
            ("par = $PAR$PAR", {"$PAR": "value"}, "par = valuevalue"),  # multiple replacements
            ("par = $PARAMETER + $PAR", {"$PARAMETER": 1, "$PAR": "value"}, "par = 1 + value"),  # test word boundaries
            ("par = $PARAMETER + $PAR", {"$PAR": 1, "$PARAMETER": "value"}, "par = value + 1"),
            ("par = $PAR + $PARAMETER + $PAR", {"$PAR": 1, "$PARAMETER": "value"}, "par = 1 + value + 1")
        ]

        for source, replacements, expected_source in cases:
            with self.subTest(source=source):
                self._awg_module.getInt.side_effect = [0, 0]  # successful compilation; upload completed
                self._awg_module.getDouble.side_effect = [1.0]  # upload progress

                self._hdawg.compile_and_upload(source, replacements)

                self._awg_module.set.assert_called_once_with("awgModule/compiler/sourcestring", expected_source)
                self.assertTrue(self._hdawg.compilation_successful())
                self._awg_module.set.reset_mock()

    def test_invalid_replacements(self):
        """Test replacement patterns."""
        with self.assertRaises(NameError):
            self._hdawg.compile_and_upload("", {"$1": 0})

        with self.assertRaises(NameError):
            self._hdawg.compile_and_upload("", {"$_": 0})

        with self.assertRaises(NameError):
            self._hdawg.compile_and_upload("", {"$_1": 0})

    def test_incomplete_replacements(self):
        """Test replacement patterns."""
        with self.assertRaises(KeyError):
            self._hdawg.compile_and_upload("$PAR1 $PAR2", {"$PAR1": 0})

    def test_replacement_value_good_types(self):
        """Test replacement type."""
        source = "par = $PAR"
        cases = [
            ("value", "par = value"),
            (1, "par = 1"),
            (3.14, "par = 3.14"),
        ]

        for value, expected_source in cases:
            with self.subTest(value=value):
                self._awg_module.getInt.side_effect = [0, 0]  # successful compilation; upload completed
                self._awg_module.getDouble.side_effect = [1.0]  # upload progress

                self._hdawg.compile_and_upload(source, {"$PAR": value})

                self._awg_module.set.assert_called_once_with("awgModule/compiler/sourcestring", expected_source)
                self.assertTrue(self._hdawg.compilation_successful())
                self._awg_module.set.reset_mock()

    def test_replacement_value_bad_types(self):
        """Test replacement type."""
        with self.assertRaises(ValueError):
            self._hdawg.compile_and_upload("", {"$PAR": None})  # wrong type

        with self.assertRaises(ValueError):
            self._hdawg.compile_and_upload("", {"$PAR": object()})  # wrong type

    def test_compile_error(self):
        """Test failed compile sequence."""
        source = "while(true) {}"

        self._awg_module.getInt.side_effect = [1]  # failed compilation

        expected_calls = [
            call.set("awgModule/compiler/sourcestring", source),
            call.getInt("awgModule/compiler/status"),
            call.getString("awgModule/compiler/statusstring")
        ]

        self._hdawg.compile_and_upload(source)

        self.assertEqual(self._awg_module.mock_calls, expected_calls)
        self.assertFalse(self._hdawg.compilation_successful())

    def test_compile_timeout(self):
        """Test timed-out compile sequence."""
        source = "while(true) {}"

        self._awg_module.getInt.return_value = -1  # idle

        expected_calls = [
            call.set("awgModule/compiler/sourcestring", source),
            call.getInt("awgModule/compiler/status"),
            call.getInt("awgModule/compiler/status"),
        ]

        with self.assertRaises(RuntimeError):
            self._hdawg.compile_and_upload(source)

        self.assertEqual(self._awg_module.mock_calls, expected_calls)

    def test_elf_upload_error(self):
        """Test failed ELF upload sequence."""
        source = "while(true) {}"

        self._awg_module.getInt.side_effect = [0, 1]  # successful compilation; failed upload
        self._awg_module.getDouble.side_effect = [0.3]  # upload progress

        expected_calls = [
            call.set("awgModule/compiler/sourcestring", source),
            call.getInt("awgModule/compiler/status"),
            call.getDouble("awgModule/progress"),
            call.getInt("awgModule/elf/status"),
        ]

        self._hdawg.compile_and_upload(source)

        self.assertEqual(self._awg_module.mock_calls, expected_calls)
        self.assertFalse(self._hdawg.compilation_successful())

    def test_elf_upload_timeout(self):
        """Test timed-out ELF upload sequence."""
        source = "while(true) {}"

        self._awg_module.getInt.side_effect = [0, -1, -1]  # successful compilation; upload timed out
        self._awg_module.getDouble.side_effect = [0.0, 0.0]  # upload progress

        expected_calls = [
            call.set("awgModule/compiler/sourcestring", source),
            call.getInt("awgModule/compiler/status"),
            call.getDouble("awgModule/progress"),
            call.getInt("awgModule/elf/status"),
            call.getDouble("awgModule/progress"),
            call.getInt("awgModule/elf/status"),
        ]

        with self.assertRaises(RuntimeError):
            self._hdawg.compile_and_upload(source)

        self.assertEqual(self._awg_module.mock_calls, expected_calls)

    def test_upload_waveform(self):
        """Test waveform upload."""
        wave1 = [1, 2, 3]
        wave2 = [4, 5, 6]
        markers = [7, 8, 9]

        hdawg.zhinst.utils.convert_awg_waveform = Mock()
        hdawg.zhinst.utils.convert_awg_waveform.return_value = "WAVEFORM"

        self._hdawg.upload_waveform(0, 0, wave1, wave2, markers)

        expected_utils_calls = [
            call.convert_awg_waveform(wave1, wave2, markers)
        ]
        hdawg.zhinst.utils.assert_has_calls(expected_utils_calls)

        expected_daq_server_calls = [
            call.setVector(f"/{_DEVICE_NAME}/awgs/0/waveform/waves/0", "WAVEFORM")
        ]
        self._daq_server.assert_has_calls(expected_daq_server_calls)

    def test_upload_waveforms_small_batch(self):
        """Test uploading a small batch of waveforms (less than batch size limit)."""
        unpacked_waveforms = []
        batch_size = 10
        for i in range(batch_size):
            awg_index = i % 4
            waveform_index = i
            wave1 = np.array([i + 1, i + 2, i + 3])
            wave2 = np.array([i + 4, i + 5, i + 6]) if i % 2 else None
            markers = np.array([i + 7, i + 8, i + 9]) if i % 2 else None
            unpacked_waveforms.append((awg_index, waveform_index, wave1, wave2, markers))

        hdawg.zhinst.utils.convert_awg_waveform = Mock()
        hdawg.zhinst.utils.convert_awg_waveform.return_value = "WAVEFORM"

        self._hdawg.upload_waveforms(unpacked_waveforms)

        expected_utils_calls = [
            call.convert_awg_waveform(unpacked_waveforms[i][:3]) for i in range(batch_size)
        ]
        [hdawg.zhinst.utils.assert_has_calls(c[2]) for c in expected_utils_calls]

        expected_daq_server_calls = [
            call.setVector(f"/{_DEVICE_NAME}/awgs/{i%4}/waveform/waves/{i}", "WAVEFORM") for i in range(batch_size)
        ]
        [self._daq_server.assert_has_calls(c[2]) for c in expected_daq_server_calls]
        self._daq_server.set.assert_called_once()

    def test_upload_waveforms_large_batch(self):
        """Test uploading a 'large' batch of waveforms (more than batch size limit but exact multiple of)."""
        unpacked_waveforms = []
        batch_size = 30
        for i in range(batch_size):
            awg_index = i % 4
            waveform_index = i
            wave1 = np.array([i + 1, i + 2, i + 3])
            wave2 = np.array([i + 4, i + 5, i + 6])
            markers = np.array([i + 7, i + 8, i + 9])
            unpacked_waveforms.append((awg_index, waveform_index, wave1, wave2, markers))

        hdawg.zhinst.utils.convert_awg_waveform = Mock()
        hdawg.zhinst.utils.convert_awg_waveform.return_value = "WAVEFORM"

        self._hdawg.upload_waveforms(unpacked_waveforms, batch_size=batch_size // 3)

        expected_utils_calls = [
            call.convert_awg_waveform(unpacked_waveforms[i][:3]) for i in range(batch_size)
        ]
        [hdawg.zhinst.utils.assert_has_calls(c[2]) for c in expected_utils_calls]

        expected_daq_server_calls = [
            call.setVector(f"/{_DEVICE_NAME}/awgs/{i%4}/waveform/waves/{i}", "WAVEFORM") for i in range(batch_size)
        ]
        [self._daq_server.assert_has_calls(c[2]) for c in expected_daq_server_calls]
        self.assertEqual(self._daq_server.set.call_count, batch_size / 10)

    def test_upload_waveforms_large_batch_2(self):
        """Test uploading a 'large' batch of waveforms (more than batch size limit and not exact multiple of)."""
        unpacked_waveforms = []
        batch_size = 31
        for i in range(batch_size):
            awg_index = i % 4
            waveform_index = i
            wave1 = np.array([i + 1, i + 2, i + 3])
            wave2 = np.array([i + 4, i + 5, i + 6])
            markers = np.array([i + 7, i + 8, i + 9])
            unpacked_waveforms.append((awg_index, waveform_index, wave1, wave2, markers))

        hdawg.zhinst.utils.convert_awg_waveform = Mock()
        hdawg.zhinst.utils.convert_awg_waveform.return_value = "WAVEFORM"

        self._hdawg.upload_waveforms(unpacked_waveforms, batch_size=batch_size // 3)

        expected_utils_calls = [
            call.convert_awg_waveform(unpacked_waveforms[i][:3]) for i in range(batch_size)
        ]
        [hdawg.zhinst.utils.assert_has_calls(c[2]) for c in expected_utils_calls]

        expected_daq_server_calls = [
            call.setVector(f"/{_DEVICE_NAME}/awgs/{i%4}/waveform/waves/{i}", "WAVEFORM") for i in range(batch_size)
        ]
        [self._daq_server.assert_has_calls(c[2]) for c in expected_daq_server_calls]
        self.assertEqual(self._daq_server.set.call_count, np.ceil(batch_size / 10))

    def test_upload_command_table(self):
        """Test command table upload."""
        command_table_entries = [
            {
                "index": 1,
                "waveform": {
                    "playZero": True,
                    "length": 128
                }
            }
        ]

        self._hdawg.upload_command_table(0, command_table_entries)

        expected_daq_server_calls = [
            call.setVector(f"/{_DEVICE_NAME}/awgs/0/commandtable/data", ANY)
        ]
        self._daq_server.assert_has_calls(expected_daq_server_calls)

        uploaded_ct = json.loads(self._daq_server.setVector.call_args[0][1])
        self.assertListEqual(uploaded_ct["table"], command_table_entries)

    def test_upload_empty_command_table(self):
        """Test command table upload."""
        table = []
        awg_index = 1
        # Create the command table from the provided entries.
        command_table = {
            "$schema": "http://docs.zhinst.com/hdawg/commandtable/v2/schema",
            "header": {
                "version": "0.2"
            },
            "table": table
        }
        set_vector = "/{}/awgs/{}/commandtable/data".format(_DEVICE_NAME, awg_index)

        self._hdawg.upload_command_table(awg_index, table)
        self._daq_server.setVector.assert_called_once_with(set_vector, json.dumps(command_table, allow_nan=False))

    def test_upload_invalid_command_table(self):
        """Test invalid command table upload."""
        command_table_entries = [
            {
                "index": 1,
                "waveform": {
                    "playZero": True,
                    "length": 123  # must be multiple of 16
                }
            }
        ]

        with self.assertRaises(ValueError):
            self._hdawg.upload_command_table(0, command_table_entries)

    def test_upload_command_table_wrong_awg(self):
        """Test invalid command table upload."""
        with self.assertRaises(ValueError):
            self._hdawg.upload_command_table(-1, [])

        with self.assertRaises(ValueError):
            self._hdawg.upload_command_table(4, [])

    def test_sync(self):
        """Test sync method."""
        self._hdawg.sync()

        expected_daq_server_calls = [
            call.sync()
        ]
        self.assertEqual(self._daq_server.mock_calls, expected_daq_server_calls)

    def test_set_channel_grouping(self):
        """Test channel grouping setting."""
        self._hdawg.set_channel_grouping(2)
        self._check_set_value_int("system/awg/channelgrouping", 2)

        with self.assertRaises(ValueError):
            self._hdawg.set_channel_grouping(1)

    def test_set_reference_clock_source(self):
        """Test reference clock source setting."""
        self._hdawg.set_reference_clock_source(0)
        self._check_set_value_int("system/clocks/referenceclock/source", 0)
        self._hdawg.set_reference_clock_source(1)
        self._check_set_value_int("system/clocks/referenceclock/source", 1)

        with self.assertRaises(ValueError):
            self._hdawg.set_reference_clock_source(2)

    def test_get_reference_clock_status(self):
        """Test reference clock source status."""
        self._hdawg.get_reference_clock_status()
        self._check_get_value_int("system/clocks/referenceclock/status")

    def test_set_sample_clock_frequency(self):
        """"Test sample clock setting."""
        self._hdawg.set_sample_clock_frequency(1234.5e6)
        self._check_set_value_double("system/clocks/sampleclock/freq", 1234.5e6)

    def test_get_sample_clock_status(self):
        """Test reference clock source status."""
        self._hdawg.get_sample_clock_status()
        self._check_get_value_int("system/clocks/sampleclock/status")

    def test_set_trigger_impedance(self):
        """Test trigger impedance setting."""
        self._hdawg.set_trigger_impedance(0, 0)
        self._check_set_value_int("triggers/in/0/imp50", 0)
        self._hdawg.set_trigger_impedance(0, 1)
        self._check_set_value_int("triggers/in/0/imp50", 1)
        self._hdawg.set_trigger_impedance(7, 0)
        self._check_set_value_int("triggers/in/7/imp50", 0)
        self._hdawg.set_trigger_impedance(7, 1)
        self._check_set_value_int("triggers/in/7/imp50", 1)

        with self.assertRaises(ValueError):
            self._hdawg.set_trigger_impedance(-1, 0)

        with self.assertRaises(ValueError):
            self._hdawg.set_trigger_impedance(8, 0)

        with self.assertRaises(ValueError):
            self._hdawg.set_trigger_impedance(0, 2)

        with self.assertRaises(ValueError):
            self._hdawg.set_trigger_impedance(8, 2)

    def test_set_trigger_level(self):
        """Test trigger level setting."""
        self._hdawg.set_trigger_level(0, 0.0)
        self._check_set_value_double("triggers/in/0/level", 0.0)
        self._hdawg.set_trigger_level(0, -5.0)
        self._check_set_value_double("triggers/in/0/level", -5.0)
        self._hdawg.set_trigger_level(0, 5.0)
        self._check_set_value_double("triggers/in/0/level", 5.0)
        self._hdawg.set_trigger_level(7, 0.0)
        self._check_set_value_double("triggers/in/7/level", 0.0)
        self._hdawg.set_trigger_level(7, -5.0)
        self._check_set_value_double("triggers/in/7/level", -5.0)
        self._hdawg.set_trigger_level(7, 5.0)
        self._check_set_value_double("triggers/in/7/level", 5.0)

        with self.assertRaises(ValueError):
            self._hdawg.set_trigger_level(-1, 0)

        with self.assertRaises(ValueError):
            self._hdawg.set_trigger_level(8, 0)

        with self.assertRaises(ValueError):
            self._hdawg.set_trigger_level(0, 10.0)

        with self.assertRaises(ValueError):
            self._hdawg.set_trigger_level(0, -10.0)

    def test_set_marker_source(self):
        """Test marker setting."""
        # NOTE: 16 is invalid value
        for i in (0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 17, 18):
            self._hdawg.set_marker_source(0, i)
            self._check_set_value_int("triggers/out/0/source", i)

        self._hdawg.set_marker_source(7, 0)
        self._check_set_value_int("triggers/out/7/source", 0)

        with self.assertRaises(ValueError):
            self._hdawg.set_marker_source(0, -1)

        with self.assertRaises(ValueError):
            self._hdawg.set_marker_source(0, 16)

        with self.assertRaises(ValueError):
            self._hdawg.set_marker_source(0, 19)

        with self.assertRaises(ValueError):
            self._hdawg.set_marker_source(-1, 0)

        with self.assertRaises(ValueError):
            self._hdawg.set_marker_source(8, 0)

    def test_set_marker_delay(self):
        """Test marker delay."""
        self._hdawg.set_marker_delay(0, 0.0)
        self._check_set_value_double("triggers/out/0/delay", 0.0)

        with self.assertRaises(ValueError):
            self._hdawg.set_marker_delay(-1, 0.0)

        with self.assertRaises(ValueError):
            self._hdawg.set_marker_delay(8, 0.0)

    def test_set_dig_trigger_source(self):
        """Test digital trigger source setting."""
        self._hdawg.set_dig_trigger_source(0, 0, 0)
        self._check_set_value_int("awgs/0/auxtriggers/0/channel", 0)
        self._hdawg.set_dig_trigger_source(3, 0, 0)
        self._check_set_value_int("awgs/3/auxtriggers/0/channel", 0)
        self._hdawg.set_dig_trigger_source(0, 1, 0)
        self._check_set_value_int("awgs/0/auxtriggers/1/channel", 0)
        self._hdawg.set_dig_trigger_source(0, 0, 7)
        self._check_set_value_int("awgs/0/auxtriggers/0/channel", 7)

        with self.assertRaises(ValueError):
            self._hdawg.set_dig_trigger_source(-1, 0, 0)

        with self.assertRaises(ValueError):
            self._hdawg.set_dig_trigger_source(4, 0, 0)

        with self.assertRaises(ValueError):
            self._hdawg.set_dig_trigger_source(0, -1, 0)

        with self.assertRaises(ValueError):
            self._hdawg.set_dig_trigger_source(0, 2, 0)

        with self.assertRaises(ValueError):
            self._hdawg.set_dig_trigger_source(0, 0, -1)

        with self.assertRaises(ValueError):
            self._hdawg.set_dig_trigger_source(0, 0, 8)

    def test_set_dig_trigger_slope(self):
        """Test digital trigger slope setting."""
        self._hdawg.set_dig_trigger_slope(0, 0, 0)
        self._check_set_value_int("awgs/0/auxtriggers/0/slope", 0)
        self._hdawg.set_dig_trigger_slope(3, 0, 0)
        self._check_set_value_int("awgs/3/auxtriggers/0/slope", 0)
        self._hdawg.set_dig_trigger_slope(0, 1, 0)
        self._check_set_value_int("awgs/0/auxtriggers/1/slope", 0)
        self._hdawg.set_dig_trigger_slope(0, 0, 3)
        self._check_set_value_int("awgs/0/auxtriggers/0/slope", 3)

        with self.assertRaises(ValueError):
            self._hdawg.set_dig_trigger_slope(-1, 0, 0)

        with self.assertRaises(ValueError):
            self._hdawg.set_dig_trigger_slope(4, 0, 0)

        with self.assertRaises(ValueError):
            self._hdawg.set_dig_trigger_slope(0, -1, 0)

        with self.assertRaises(ValueError):
            self._hdawg.set_dig_trigger_slope(0, 2, 0)

        with self.assertRaises(ValueError):
            self._hdawg.set_dig_trigger_slope(0, 0, -1)

        with self.assertRaises(ValueError):
            self._hdawg.set_dig_trigger_slope(0, 0, 4)

    def test_set_output_amplitude(self):
        """Test output amplitude setting."""
        self._hdawg.set_output_amplitude(0, 0, 1.0)
        self._check_set_value_double("awgs/0/outputs/0/amplitude", 1.0)
        self._hdawg.set_output_amplitude(0, 1, 1.0)
        self._check_set_value_double("awgs/0/outputs/1/amplitude", 1.0)
        self._hdawg.set_output_amplitude(0, 0, -1.0)
        self._check_set_value_double("awgs/0/outputs/0/amplitude", -1.0)
        self._hdawg.set_output_amplitude(0, 1, -1.0)
        self._check_set_value_double("awgs/0/outputs/1/amplitude", -1.0)
        self._hdawg.set_output_amplitude(3, 0, 1.0)
        self._check_set_value_double("awgs/3/outputs/0/amplitude", 1.0)
        self._hdawg.set_output_amplitude(3, 1, 1.0)
        self._check_set_value_double("awgs/3/outputs/1/amplitude", 1.0)
        self._hdawg.set_output_amplitude(3, 0, -1.0)
        self._check_set_value_double("awgs/3/outputs/0/amplitude", -1.0)
        self._hdawg.set_output_amplitude(3, 1, -1.0)
        self._check_set_value_double("awgs/3/outputs/1/amplitude", -1.0)

        with self.assertRaises(ValueError):
            self._hdawg.set_output_amplitude(-1, 0, 1.0)

        with self.assertRaises(ValueError):
            self._hdawg.set_output_amplitude(4, 0, 1.0)

        with self.assertRaises(ValueError):
            self._hdawg.set_output_amplitude(0, -1, 1.0)

        with self.assertRaises(ValueError):
            self._hdawg.set_output_amplitude(0, 3, 1.0)

    def test_get_output_amplitude(self):
        """Test output amplitude query."""
        self._hdawg.get_output_amplitude(0, 0)
        self._check_get_value_double("awgs/0/outputs/0/amplitude")
        self._hdawg.get_output_amplitude(0, 1)
        self._check_get_value_double("awgs/0/outputs/1/amplitude")
        self._hdawg.get_output_amplitude(0, 0)
        self._check_get_value_double("awgs/0/outputs/0/amplitude")
        self._hdawg.get_output_amplitude(0, 1)
        self._check_get_value_double("awgs/0/outputs/1/amplitude")
        self._hdawg.get_output_amplitude(3, 0)
        self._check_get_value_double("awgs/3/outputs/0/amplitude")
        self._hdawg.get_output_amplitude(3, 1)
        self._check_get_value_double("awgs/3/outputs/1/amplitude")
        self._hdawg.get_output_amplitude(3, 0)
        self._check_get_value_double("awgs/3/outputs/0/amplitude")
        self._hdawg.get_output_amplitude(3, 1)
        self._check_get_value_double("awgs/3/outputs/1/amplitude")

        with self.assertRaises(ValueError):
            self._hdawg.get_output_amplitude(-1, 0)

        with self.assertRaises(ValueError):
            self._hdawg.get_output_amplitude(4, 0)

        with self.assertRaises(ValueError):
            self._hdawg.get_output_amplitude(0, -1)

        with self.assertRaises(ValueError):
            self._hdawg.get_output_amplitude(0, 3)

    def test_set_output_channel_hold(self):
        """Test output hold setting."""
        self._hdawg.set_output_channel_hold(0, 0, 1)
        self._check_set_value_int("awgs/0/outputs/0/hold", 1)
        self._hdawg.set_output_channel_hold(0, 1, 1)
        self._check_set_value_int("awgs/0/outputs/1/hold", 1)
        self._hdawg.set_output_channel_hold(0, 0, 1)
        self._check_set_value_int("awgs/0/outputs/0/hold", 1)
        self._hdawg.set_output_channel_hold(0, 1, 1)
        self._check_set_value_int("awgs/0/outputs/1/hold", 1)
        self._hdawg.set_output_channel_hold(3, 0, 1)
        self._check_set_value_int("awgs/3/outputs/0/hold", 1)
        self._hdawg.set_output_channel_hold(3, 1, 1)
        self._check_set_value_int("awgs/3/outputs/1/hold", 1)
        self._hdawg.set_output_channel_hold(3, 0, 1)
        self._check_set_value_int("awgs/3/outputs/0/hold", 1)
        self._hdawg.set_output_channel_hold(3, 1, 1)
        self._check_set_value_int("awgs/3/outputs/1/hold", 1)
        self._hdawg.set_output_channel_hold(0, 0, 0)
        self._check_set_value_int("awgs/0/outputs/0/hold", 0)

        with self.assertRaises(ValueError):
            self._hdawg.set_output_channel_hold(-1, 0, 1)

        with self.assertRaises(ValueError):
            self._hdawg.set_output_channel_hold(4, 0, 1)

        with self.assertRaises(ValueError):
            self._hdawg.set_output_channel_hold(0, -1, 1)

        with self.assertRaises(ValueError):
            self._hdawg.set_output_channel_hold(0, 3, 1)

        with self.assertRaises(ValueError):
            self._hdawg.set_output_channel_hold(0, 0, -1)

        with self.assertRaises(ValueError):
            self._hdawg.set_output_channel_hold(0, 0, 2)

    def test_get_output_channel_hold(self):
        """Test output hold query."""
        self._hdawg.get_output_channel_hold(0, 0)
        self._check_get_value_int("awgs/0/outputs/0/hold")
        self._hdawg.get_output_channel_hold(0, 1)
        self._check_get_value_int("awgs/0/outputs/1/hold")
        self._hdawg.get_output_channel_hold(3, 0)
        self._check_get_value_int("awgs/3/outputs/0/hold")
        self._hdawg.get_output_channel_hold(3, 1)
        self._check_get_value_int("awgs/3/outputs/1/hold")

        with self.assertRaises(ValueError):
            self._hdawg.get_output_channel_hold(-1, 0)

        with self.assertRaises(ValueError):
            self._hdawg.get_output_channel_hold(4, 0)

        with self.assertRaises(ValueError):
            self._hdawg.get_output_channel_hold(0, -1)

        with self.assertRaises(ValueError):
            self._hdawg.get_output_channel_hold(0, 3)

    def test_get_output_channel_on(self):
        """Test output channel on/off state query."""
        self._hdawg.get_output_channel_on(0)
        self._check_get_value_int("sigouts/0/on")
        self._hdawg.get_output_channel_on(7)
        self._check_get_value_int("sigouts/7/on")

        with self.assertRaises(ValueError):
            self._hdawg.get_output_channel_on(-1)

        with self.assertRaises(ValueError):
            self._hdawg.get_output_channel_on(8)

    def test_set_output_channel_on(self):
        """Test output channel on/off switching."""
        self._hdawg.set_output_channel_on(0, 0)
        self._check_set_value_int("sigouts/0/on", 0)
        self._hdawg.set_output_channel_on(0, 1)
        self._check_set_value_int("sigouts/0/on", 1)
        self._hdawg.set_output_channel_on(7, 0)
        self._check_set_value_int("sigouts/7/on", 0)
        self._hdawg.set_output_channel_on(7, 1)
        self._check_set_value_int("sigouts/7/on", 1)

        with self.assertRaises(ValueError):
            self._hdawg.set_output_channel_on(-1, 0)

        with self.assertRaises(ValueError):
            self._hdawg.set_output_channel_on(8, 0)

        with self.assertRaises(ValueError):
            self._hdawg.set_output_channel_on(0, -1)

        with self.assertRaises(ValueError):
            self._hdawg.set_output_channel_on(0, 2)

    def test_set_output_channel_range(self):
        """Test output channel range setting."""
        self._hdawg.set_output_channel_range(0, 0.0)
        self._check_set_value_double("sigouts/0/range", 0.0)
        self._hdawg.set_output_channel_range(0, 5.0)
        self._check_set_value_double("sigouts/0/range", 5.0)
        self._hdawg.set_output_channel_range(7, 0.0)
        self._check_set_value_double("sigouts/7/range", 0.0)
        self._hdawg.set_output_channel_range(7, 5.0)
        self._check_set_value_double("sigouts/7/range", 5.0)

        with self.assertRaises(ValueError):
            self._hdawg.set_output_channel_on(-1, 0.0)

        with self.assertRaises(ValueError):
            self._hdawg.set_output_channel_on(8, 0.0)

        with self.assertRaises(ValueError):
            self._hdawg.set_output_channel_on(0, -1.0)

        with self.assertRaises(ValueError):
            self._hdawg.set_output_channel_on(0, 6.0)

    def test_set_output_channel_offset(self):
        """Test output channel offset setting."""
        self._hdawg.set_output_channel_offset(0, 0.0)
        self._check_set_value_double("sigouts/0/offset", 0.0)
        self._hdawg.set_output_channel_offset(0, 1.25)
        self._check_set_value_double("sigouts/0/offset", 1.25)
        self._hdawg.set_output_channel_offset(0, -1.25)
        self._check_set_value_double("sigouts/0/offset", -1.25)
        self._hdawg.set_output_channel_offset(7, 0.0)
        self._check_set_value_double("sigouts/7/offset", 0.0)
        self._hdawg.set_output_channel_offset(7, 1.25)
        self._check_set_value_double("sigouts/7/offset", 1.25)
        self._hdawg.set_output_channel_offset(7, -1.25)
        self._check_set_value_double("sigouts/7/offset", -1.25)

        with self.assertRaises(ValueError):
            self._hdawg.set_output_channel_on(-1, 0.0)

        with self.assertRaises(ValueError):
            self._hdawg.set_output_channel_on(8, 0.0)

        with self.assertRaises(ValueError):
            self._hdawg.set_output_channel_on(0, -2.0)

        with self.assertRaises(ValueError):
            self._hdawg.set_output_channel_on(0, 2.0)

    def test_get_output_channel_delay(self):
        """Test getting output channel delay."""
        self._hdawg.get_output_channel_delay(0)
        self._check_get_value_float("sigouts/0/delay")
        self._hdawg.get_output_channel_delay(7)
        self._check_get_value_float("sigouts/7/delay")

        with self.assertRaises(ValueError):
            self._hdawg.get_output_channel_delay(-1)

        with self.assertRaises(ValueError):
            self._hdawg.get_output_channel_delay(8)

    def test_set_output_channel_delay(self):
        """Test output channel delay setting."""
        self._hdawg.set_output_channel_delay(0, 0.0)
        self._check_set_value_double("sigouts/0/delay", 0.0)
        self._hdawg.set_output_channel_delay(0, 25e-9)
        self._check_set_value_double("sigouts/0/delay", 25e-9)
        self._hdawg.set_output_channel_delay(7, 0.0)
        self._check_set_value_double("sigouts/7/delay", 0.0)
        self._hdawg.set_output_channel_delay(7, 25e-9)
        self._check_set_value_double("sigouts/7/delay", 25e-9)

        with self.assertRaises(ValueError):
            self._hdawg.set_output_channel_delay(-1, 0.0)

        with self.assertRaises(ValueError):
            self._hdawg.set_output_channel_delay(8, 0.0)

        with self.assertRaises(ValueError):
            self._hdawg.set_output_channel_delay(0, -1.0)

        with self.assertRaises(ValueError):
            self._hdawg.set_output_channel_delay(0, 1.0)

    def test_set_output_channel_direct(self):
        """Test output channel direct setting."""
        self._hdawg.set_output_channel_direct(0, 0)
        self._check_set_value_int("sigouts/0/direct", 0)
        self._hdawg.set_output_channel_direct(0, 1)
        self._check_set_value_int("sigouts/0/direct", 1)
        self._hdawg.set_output_channel_direct(7, 0)
        self._check_set_value_int("sigouts/7/direct", 0)
        self._hdawg.set_output_channel_direct(7, 1)
        self._check_set_value_int("sigouts/7/direct", 1)

        with self.assertRaises(ValueError):
            self._hdawg.set_output_channel_direct(-1, 0)

        with self.assertRaises(ValueError):
            self._hdawg.set_output_channel_direct(8, 0)

        with self.assertRaises(ValueError):
            self._hdawg.set_output_channel_direct(0, -1)

        with self.assertRaises(ValueError):
            self._hdawg.set_output_channel_direct(0, 2)

    def test_set_output_channel_filter(self):
        """Test output channel filter setting."""
        self._hdawg.set_output_channel_filter(0, 0)
        self._check_set_value_int("sigouts/0/filter", 0)
        self._hdawg.set_output_channel_filter(0, 1)
        self._check_set_value_int("sigouts/0/filter", 1)
        self._hdawg.set_output_channel_filter(7, 0)
        self._check_set_value_int("sigouts/7/filter", 0)
        self._hdawg.set_output_channel_filter(7, 1)
        self._check_set_value_int("sigouts/7/filter", 1)

        with self.assertRaises(ValueError):
            self._hdawg.set_output_channel_filter(-1, 0)

        with self.assertRaises(ValueError):
            self._hdawg.set_output_channel_filter(8, 0)

        with self.assertRaises(ValueError):
            self._hdawg.set_output_channel_filter(0, -1)

        with self.assertRaises(ValueError):
            self._hdawg.set_output_channel_filter(0, 2)

    def test_set_dio_mode(self):
        """Test DIO mode setting."""
        for i in range(4):
            self._hdawg.set_dio_mode(i)
            self._check_set_value_int("dios/0/mode", i)

        with self.assertRaises(ValueError):
            self._hdawg.set_dio_mode(-1)

        with self.assertRaises(ValueError):
            self._hdawg.set_dio_mode(4)

    def test_set_dio_drive(self):
        """Test DIO drive setting."""
        for i in range(16):
            self._hdawg.set_dio_drive(i)
            self._check_set_value_int("dios/0/drive", i)

        with self.assertRaises(ValueError):
            self._hdawg.set_dio_drive(-1)

        with self.assertRaises(ValueError):
            self._hdawg.set_dio_drive(16)

    def test_set_dio_strobe_index(self):
        """Test DIO strobe index setting."""
        self._hdawg.set_dio_strobe_index(0, 0)
        self._check_set_value_int("awgs/0/dio/strobe/index", 0)
        self._hdawg.set_dio_strobe_index(0, 31)
        self._check_set_value_int("awgs/0/dio/strobe/index", 31)
        self._hdawg.set_dio_strobe_index(3, 0)
        self._check_set_value_int("awgs/3/dio/strobe/index", 0)
        self._hdawg.set_dio_strobe_index(3, 31)
        self._check_set_value_int("awgs/3/dio/strobe/index", 31)

        with self.assertRaises(ValueError):
            self._hdawg.set_dio_strobe_index(-1, 0)

        with self.assertRaises(ValueError):
            self._hdawg.set_dio_strobe_index(4, 0)

        with self.assertRaises(ValueError):
            self._hdawg.set_dio_strobe_index(0, -1)

        with self.assertRaises(ValueError):
            self._hdawg.set_dio_strobe_index(0, 32)

    def test_set_dio_strobe_slope(self):
        """Test DIO strobe slope setting."""
        self._hdawg.set_dio_strobe_slope(0, 0)
        self._check_set_value_int("awgs/0/dio/strobe/slope", 0)
        self._hdawg.set_dio_strobe_slope(0, 3)
        self._check_set_value_int("awgs/0/dio/strobe/slope", 3)
        self._hdawg.set_dio_strobe_slope(3, 0)
        self._check_set_value_int("awgs/3/dio/strobe/slope", 0)
        self._hdawg.set_dio_strobe_slope(3, 3)
        self._check_set_value_int("awgs/3/dio/strobe/slope", 3)

        with self.assertRaises(ValueError):
            self._hdawg.set_dio_strobe_slope(-1, 0)

        with self.assertRaises(ValueError):
            self._hdawg.set_dio_strobe_slope(4, 0)

        with self.assertRaises(ValueError):
            self._hdawg.set_dio_strobe_slope(0, -1)

        with self.assertRaises(ValueError):
            self._hdawg.set_dio_strobe_slope(0, 4)

    def test_get_user_register(self):
        """Test get user register."""
        self._hdawg.get_user_register(0, 0)
        self._check_get_value_int("awgs/0/userregs/0")
        self._hdawg.get_user_register(0, 15)
        self._check_get_value_int("awgs/0/userregs/15")
        self._hdawg.get_user_register(3, 0)
        self._check_get_value_int("awgs/3/userregs/0")
        self._hdawg.get_user_register(3, 15)
        self._check_get_value_int("awgs/3/userregs/15")

        with self.assertRaises(ValueError):
            self._hdawg.get_user_register(-1, 0)

        with self.assertRaises(ValueError):
            self._hdawg.get_user_register(4, 0)

        with self.assertRaises(ValueError):
            self._hdawg.get_user_register(0, -1)

        with self.assertRaises(ValueError):
            self._hdawg.get_user_register(0, 16)

    def test_set_user_register(self):
        """Test user register setting."""
        self._hdawg.set_user_register(0, 0, 123)
        self._check_set_value_int("awgs/0/userregs/0", 123)
        self._hdawg.set_user_register(0, 15, -123)
        self._check_set_value_int("awgs/0/userregs/15", -123)
        self._hdawg.set_user_register(3, 0, 123)
        self._check_set_value_int("awgs/3/userregs/0", 123)
        self._hdawg.set_user_register(3, 15, -123)
        self._check_set_value_int("awgs/3/userregs/15", -123)

        with self.assertRaises(ValueError):
            self._hdawg.set_user_register(-1, 0, 456)

        with self.assertRaises(ValueError):
            self._hdawg.set_user_register(4, 0, 456)

        with self.assertRaises(ValueError):
            self._hdawg.set_user_register(0, -1, 789)

        with self.assertRaises(ValueError):
            self._hdawg.set_user_register(0, 16, 789)

    def test_awg_module_enabled(self):
        """Test AWG enable on/off."""
        self._hdawg.get_awg_module_enabled(0)
        self._check_get_value_int("awgs/0/enable")
        self._hdawg.get_awg_module_enabled(3)
        self._check_get_value_int("awgs/3/enable")

        with self.assertRaises(ValueError):
            self._hdawg.get_awg_module_enabled(-1)

        with self.assertRaises(ValueError):
            self._hdawg.get_awg_module_enabled(4)

        self._hdawg.set_awg_module_enabled(0)
        self._hdawg.set_awg_module_enabled(1)

        expected_awg_module_calls = [
            call.set("awgModule/awg/enable", 0),
            call.set("awgModule/awg/enable", 1)
        ]
        self.assertEqual(self._awg_module.mock_calls, expected_awg_module_calls)

        with self.assertRaises(ValueError):
            self._hdawg.set_awg_module_enabled(-1)

        with self.assertRaises(ValueError):
            self._hdawg.set_awg_module_enabled(2)


if __name__ == '__main__':
    unittest.main()



"""
Library wrapper module that utilizes a Python wrapper around the PicoQuant library header files
for implemented models. The wrapper uses the manufacturer provided software libraries. The library files and
the respective licence terms can be found in the dedicated instrument software packages at PicoQuant website:
https://www.picoquant.com/products/category/tcspc-and-time-tagging-modules.
"""
import ctypes
import logging
import sys
from typing import Any

from qmi.core.exceptions import QMI_InstrumentException
from qmi.instruments.picoquant.support._hhlib_function_signatures import _hhlib_function_signatures
from qmi.instruments.picoquant.support._mhlib_function_signatures import _mhlib_function_signatures
from qmi.instruments.picoquant.support._phlib_function_signatures import _phlib_function_signatures

_logger = logging.getLogger(__name__)


class _LibWrapper:

    def __init__(self, model: str) -> None:
        """ A Library wrapper class that makes a wrapper around the PicoQuant libraries for implemented models.

        Parameters:
            model: The PicoQuant xxHarp model as a string. Possible values MH, HH, PH, TH260.

        Raises:
            FileNotFoundError: When the library file of a specific xxHarp model could not be found (not installed?)
            OSError: Trying to run on an unsupported platform, e.g. MacOS ("darwin").
        """
        self._prefix = model
        if sys.platform.startswith("linux"):
            if model == "HH":
                self._lib = ctypes.cdll.LoadLibrary("libhh400.so")
                self.annotate_function_signatures(_hhlib_function_signatures)

            elif model == "MH":
                self._lib = ctypes.cdll.LoadLibrary("libmh150.so")
                self.annotate_function_signatures(_mhlib_function_signatures)

            elif model == "PH":
                self._lib = ctypes.cdll.LoadLibrary("libph300.so")
                self.annotate_function_signatures(_phlib_function_signatures)

            else:
                raise FileNotFoundError("Unknown library: {lib}.".format(lib=model))

        elif sys.platform.startswith("win"):
            if model == "HH":
                try:
                    self._lib = ctypes.WinDLL("hhlib64.dll")  # 64-bit version
                except:
                    self._lib = ctypes.WinDLL("hhlib.dll")  # 32-bit version

                self.annotate_function_signatures(_hhlib_function_signatures)

            elif model == "MH":
                try:
                    self._lib = ctypes.WinDLL("mhlib64.dll")  # 64-bit version
                except:
                    self._lib = ctypes.WinDLL("mhlib.dll")  # 32-bit version

                self.annotate_function_signatures(_mhlib_function_signatures)

            elif model == "PH":
                try:
                    self._lib = ctypes.WinDLL("phlib64.dll")  # 64-bit version
                except:
                    self._lib = ctypes.WinDLL("phlib.dll")  # 32-bit version

                self.annotate_function_signatures(_phlib_function_signatures)

            else:
                raise FileNotFoundError("Unknown library: {lib}.".format(lib=model))

        else:
            raise OSError("Unsupported platform.")

    def __getattr__(self, item: str) -> Any:
        attr_name = "{prefix}_{item}".format(prefix=self._prefix, item=item)
        attr = getattr(self._lib, attr_name)

        def wrap_fun(*args, **kwargs):
            _logger.debug(
                "Using %s library to call function %s with arguments %s, %s",
                self._prefix, attr_name, args, kwargs
            )
            errcode = attr(*args, **kwargs)
            if errcode != 0:
                _logger.error("[%s] Call to function %s failed with errorcode %i", self._prefix, attr_name, errcode)
                raise QMI_InstrumentException(f"Interaction with PicoQuant library failed, errorcode [{errcode}].")

        return wrap_fun

    def annotate_function_signatures(self, sigs) -> None:
        """Annotate functions present in the MultiHarp shared library according to their function signatures."""
        function_signatures = sigs

        for (name, restype, argtypes) in function_signatures:
            try:
                func = getattr(self._lib, name)
                func.restype = restype
                func.argtypes = [argtype for (argname, argtype) in argtypes]
            except AttributeError:
                # Ignore functions that cannot be found.
                pass



import sys
import unittest, unittest.mock
sys.modules["usb.core"] = unittest.mock.Mock()
sys.modules["usb.core.find"] = unittest.mock.Mock()
from typing import NamedTuple
import time

from qmi.instruments.thorlabs import SensorInfo
from qmi.instruments.thorlabs.pm100d import Thorlabs_PM10x
from qmi.core.exceptions import QMI_InstrumentException
from qmi.core.context import QMI_Context

UsbDevice = NamedTuple('UsbDeviceInfo',
                        [('idProduct', hex),
                         ('idVendor', hex),
                         ('serial_number', str)])


class TestThorlabsPM10x(unittest.TestCase):

    def setUp(self) -> None:
        qmi_context = unittest.mock.MagicMock(spec=QMI_Context)
        qmi_context.name = "mock_context"

        self.vendor_id = "0x1313"
        self.product = "PM16_120"
        self.product_id = "0x807b"
        self.serialnr = "P0024208"
        self.transport_id = f"usbtmc:vendorid={self.vendor_id}:productid={self.product_id}:serialnr={self.serialnr}"
        with unittest.mock.patch("qmi.instruments.thorlabs.pm100d.create_transport") as self._transport:
            self.thorlabs = Thorlabs_PM10x(qmi_context, "powermeter", self.transport_id)

    def test_open_close(self):
        self.thorlabs.open()
        self.thorlabs.close()
        self._transport.assert_called_once_with(self.transport_id)
        self.thorlabs._transport.open.assert_called_once_with()
        self.thorlabs._transport.close.assert_called_once_with()

    @unittest.mock.patch("qmi.core.scpi_protocol.ScpiProtocol.ask")
    def test_get_idn(self, mock_read):
        """See that the get_idn method returns the expected QMI_InstrumentIdentification object"""
        mock_read.return_value = f"vendor={self.vendor_id},model={self.product},serial={self.serialnr},version=1.2.3"
        expected_vendor = f"vendor={self.vendor_id}"
        expected_model = f"model={self.product}"
        expected_serial = f"serial={self.serialnr}"
        expected_version = "version=1.2.3"

        self.thorlabs.open()
        msg = self.thorlabs.get_idn()
        self.thorlabs.close()

        self.assertEqual(msg.vendor, expected_vendor)
        self.assertEqual(msg.model, expected_model)
        self.assertEqual(msg.serial, expected_serial)
        self.assertEqual(msg.version, expected_version)

    @unittest.mock.patch("qmi.core.scpi_protocol.ScpiProtocol.ask")
    def test_get_idn_error(self, mock_read):
        """See that the get_idn method raises an exception at wrong number of returned words"""
        mock_read.return_value = f"vendor={self.vendor_id},model={self.product},serial={self.serialnr}"

        with self.assertRaises(QMI_InstrumentException):
            self.thorlabs.open()
            self.thorlabs.get_idn()

        self.thorlabs.close()

    @unittest.mock.patch("qmi.core.scpi_protocol.ScpiProtocol.ask")
    def test_reset(self, mock_read):
        """See that the reset method calls 'ask' twice"""
        mock_read.side_effect = ["OK", '+0,"No error"']

        self.thorlabs.open()
        self.thorlabs.reset()
        self.thorlabs.close()

        self.assertEqual(2, mock_read.call_count)

    @unittest.mock.patch("qmi.core.scpi_protocol.ScpiProtocol.ask")
    def test_reset_error(self, mock_read):
        """See that the reset method raises an error at error check"""
        mock_read.side_effect = ["NOK", '+1,"Some error"']

        with self.assertRaises(QMI_InstrumentException):
            self.thorlabs.open()
            self.thorlabs.reset()

        self.thorlabs.close()

    @unittest.mock.patch("qmi.core.scpi_protocol.ScpiProtocol.ask")
    def test_get_sensor_info(self, mock_read):
        """See that a SensorInfo object is returned."""
        expected = "SYST:SENS:IDN?"
        mock_read.return_value = f"name=powermeter,serial={self.serialnr},cal_msg=info,type=dunno,subtype=huh,3"

        self.thorlabs.open()
        msg = self.thorlabs.get_sensor_info()
        self.thorlabs.close()

        self.assertEqual(type(msg), SensorInfo)
        mock_read.assert_called_once_with(expected)

    @unittest.mock.patch("qmi.core.scpi_protocol.ScpiProtocol.ask")
    def test_get_sensor_info_error(self, mock_read):
        """See that an exception is raised with wrong number of returned words"""
        mock_read.return_value = f"name=powermeter,serial={self.serialnr},cal_msg=info,type=dunno,3"

        with self.assertRaises(QMI_InstrumentException):
            self.thorlabs.open()
            self.thorlabs.get_sensor_info()

        self.thorlabs.close()

    @unittest.mock.patch("qmi.core.scpi_protocol.ScpiProtocol.ask")
    def test_get_timestamped_power(self, mock_read):
        """See that a power reading with a timestamp is returned"""
        ref_timestamp = time.time()
        power = 100.0
        expected = "MEAS:POW?"
        mock_read.return_value = str(power)

        self.thorlabs.open()
        timestamp, msg = self.thorlabs.get_timestamped_power()
        self.thorlabs.close()

        self.assertEqual(msg, power)
        self.assertGreaterEqual(timestamp, ref_timestamp)
        mock_read.assert_called_once_with(expected)

    @unittest.mock.patch("qmi.core.scpi_protocol.ScpiProtocol.ask")
    def test_get_power(self, mock_read):
        """Test that only power reading is returned"""
        power = 100.0
        expected = "MEAS:POW?"
        mock_read.return_value = str(power)

        self.thorlabs.open()
        msg = self.thorlabs.get_power()
        self.thorlabs.close()

        self.assertEqual(msg, power)
        mock_read.assert_called_once_with(expected)

    @unittest.mock.patch("qmi.core.scpi_protocol.ScpiProtocol.ask")
    def test_get_range(self, mock_read):
        """Test that range value is returned."""
        range = 9.99
        expected = "SENS:POW:RANG?"
        mock_read.return_value = str(range)

        self.thorlabs.open()
        msg = self.thorlabs.get_range()
        self.thorlabs.close()

        self.assertEqual(msg, range)
        mock_read.assert_called_once_with(expected)

    @unittest.mock.patch("qmi.core.scpi_protocol.ScpiProtocol.ask")
    def test_get_range_in_use(self, mock_read):
        """Test that range value is returned."""
        range = 9.99
        expected = "SENS:POW:RANG?"
        mock_read.return_value = str(range)

        self.thorlabs.open()
        msg = self.thorlabs.get_range_in_use()
        self.thorlabs.close()

        self.assertEqual(msg, range)
        mock_read.assert_called_once_with(expected)

    @unittest.mock.patch("qmi.core.scpi_protocol.ScpiProtocol.write")
    @unittest.mock.patch("qmi.core.scpi_protocol.ScpiProtocol.ask")
    def test_set_range(self, mock_read, mock_write):
        """Test that range value can be set."""
        range = 10.0
        expected = f"SENS:POW:RANG {range}"
        mock_read.return_value = '+0,"No error"'

        self.thorlabs.open()
        self.thorlabs.set_range(range)
        self.thorlabs.close()

        mock_write.assert_called_once_with(expected)

    @unittest.mock.patch("qmi.core.scpi_protocol.ScpiProtocol.ask")
    def test_get_autorange(self, mock_read):
        """Test that autorange value is returned."""
        expected = "SENS:POW:RANG:AUTO?"
        mock_read.return_value = True

        self.thorlabs.open()
        msg = self.thorlabs.get_autorange()
        self.thorlabs.close()

        self.assertTrue(msg)
        mock_read.assert_called_once_with(expected)

    @unittest.mock.patch("qmi.core.scpi_protocol.ScpiProtocol.write")
    @unittest.mock.patch("qmi.core.scpi_protocol.ScpiProtocol.ask")
    def test_set_autorange(self, mock_read, mock_write):
        """Test that autorange can be set."""
        autorange = False
        expected = "SENS:POW:RANG:AUTO {}".format(1 if autorange else 0)
        mock_read.return_value = '+0,"No error"'

        self.thorlabs.open()
        self.thorlabs.set_autorange(autorange)
        self.thorlabs.close()

        mock_write.assert_called_once_with(expected)

    @unittest.mock.patch("qmi.core.scpi_protocol.ScpiProtocol.ask")
    def test_get_wavelength(self, mock_read):
        """Test that wavelength value is returned."""
        wavelength = 564.7
        expected = "SENS:CORR:WAV?"
        mock_read.return_value = str(wavelength)

        self.thorlabs.open()
        msg = self.thorlabs.get_wavelength()
        self.thorlabs.close()

        self.assertEqual(msg, wavelength)
        mock_read.assert_called_once_with(expected)

    @unittest.mock.patch("qmi.core.scpi_protocol.ScpiProtocol.write")
    @unittest.mock.patch("qmi.core.scpi_protocol.ScpiProtocol.ask")
    def test_set_wavelength(self, mock_read, mock_write):
        """Test that range value is returned."""
        wavelength = 563.2
        expected = "SENS:CORR:WAV {}".format(wavelength)
        mock_read.return_value = '+0,"No error"'

        self.thorlabs.open()
        self.thorlabs.set_wavelength(wavelength)
        self.thorlabs.close()

        mock_write.assert_called_once_with(expected)

    @unittest.mock.patch("qmi.core.scpi_protocol.ScpiProtocol.ask")
    def test_get_current(self, mock_read):
        """Test that current value is returned."""
        current = 1.2
        expected = "MEAS:CURR?"
        mock_read.return_value = str(current)

        self.thorlabs.open()
        msg = self.thorlabs.get_current()
        self.thorlabs.close()

        self.assertEqual(msg, current)
        mock_read.assert_called_once_with(expected)


if __name__ == '__main__':
    unittest.main()



import unittest, unittest.mock
from typing import cast

from qmi.instruments.thorlabs import Thorlabs_Tsp01
from qmi.core.transport_usbtmc_visa import QMI_VisaUsbTmcTransport
import qmi.core.exceptions
from qmi.utils.context_managers import open_close


class TestThorlabsTsp01(unittest.TestCase):
    def setUp(self):
        qmi.start("TestTsp01Context")
        self._transport_mock = unittest.mock.MagicMock(spec=QMI_VisaUsbTmcTransport)
        with unittest.mock.patch(
                'qmi.instruments.thorlabs.tsp01.create_transport',
                return_value=self._transport_mock):
            self.instr: Thorlabs_Tsp01 = qmi.make_instrument("instr", Thorlabs_Tsp01, "transport_descriptor")
            self.instr = cast(Thorlabs_Tsp01, self.instr)

    def tearDown(self):
        qmi.stop()

    def test_open_close(self):
        """Test opening and closing the instrument"""
        self.instr.open()
        self.instr.close()

        self._transport_mock.open.assert_called_once()
        self._transport_mock.close.assert_called_once()

    def test_reset(self):
        """Test the reset method."""
        expected = [b"*RST\n", b"*OPC?\n"]
        with open_close(self.instr):
            self.instr.reset()

        self._transport_mock.write.assert_called_with(expected[1])
        self._transport_mock.write.assert_any_call(expected[0])

    def test_get_idn(self):
        """Test the get_idn method."""
        expected = b"*IDN?\n"
        expected_idn = ["Thorlabs", "TSP01", "59595", "1.23"]
        return_value_str = ",".join(expected_idn) + "\n"
        self._transport_mock.read_until.return_value = return_value_str.encode()
        with open_close(self.instr):
            idn = self.instr.get_idn()

        self._transport_mock.write.assert_called_with(expected)
        self.assertEqual(idn.vendor, expected_idn[0])
        self.assertEqual(idn.model, expected_idn[1])
        self.assertEqual(idn.serial, expected_idn[2])
        self.assertEqual(idn.version, expected_idn[3])

    def test_get_idn_excepts(self):
        """Test the get_idn method excepts if too many words are returned."""
        unexpected_idn = ["koekkoek", "Thorlabs", "TSP01", "59595", "1.23"]
        return_value_str = ",".join(unexpected_idn) + "\n"
        self._transport_mock.read_until.return_value = return_value_str.encode()
        with open_close(self.instr):
            with self.assertRaises(qmi.core.exceptions.QMI_InstrumentException):
                self.instr.get_idn()

    def test_get_errors(self):
        """Test the get_errors method. Test both no errors and errors present cases."""
        expected = b"SYST:ERR?\n"
        expected_errors_1 = []
        expected_errors_2 = ["1,Very Serious Error", "2:Even Worse Error"]

        return_value_1 = b"0,What No Errors?\n"
        return_value_2 = expected_errors_2[0].encode() + b"\n"
        return_value_3 = expected_errors_2[1].encode() + b"\n"

        # First test. The reading stops at first entry as it starts with "0," and returns empty list
        self._transport_mock.read_until.side_effect = [return_value_1]
        with open_close(self.instr):
            no_errs = self.instr.get_errors()

        self.assertListEqual(no_errs, expected_errors_1)
        self._transport_mock.write.assert_called_with(expected)

        # Second test. Reading stops at third entry only and returns a list of two errors
        self._transport_mock.read_until.side_effect = [return_value_2, return_value_3, return_value_1]
        with open_close(self.instr):
            errs = self.instr.get_errors()

        self.assertListEqual(errs, expected_errors_2)
        self._transport_mock.write.assert_called_with(expected)

    def test_get_temperature(self):
        """Test get_temperature method returns temperature value."""
        sensors = [1, 2, 3]
        expected = [f"MEAS:TEMP{i}?\n" for i in sensors]
        expected_values = [10.0, 100.0, -100.0]
        for s in range(3):
            return_value_str = str(expected_values[s]) + "\n"
            self._transport_mock.read_until.return_value = return_value_str.encode()
            with open_close(self.instr):
                t = self.instr.get_temperature(sensors[s])

            self.assertEqual(t, expected_values[s])
            self._transport_mock.write.assert_called_with(expected[s].encode())

    def test_get_temperature_excepts(self):
        """Test that get_temperature call excepts on wrong sensor number and invalid response"""
        sensors = [0, 1]
        cmd = "MEAS:TEMP1?"
        expected_errors = ["Unknown temperature sensor 0", "Unexpected response to {!r}, got {!r}".format(cmd, chr(1))]
        for s in range(2):
            return_value_str = chr(s) + "\n"
            self._transport_mock.read_until.return_value = return_value_str.encode()
            with open_close(self.instr), self.assertRaises(qmi.core.exceptions.QMI_InstrumentException) as err:
                self.instr.get_temperature(sensors[s])

            self.assertEqual(str(err.exception), expected_errors[s])

    def test_get_humidity(self):
        """Test that get_humidity call returns a value."""
        expected = b"MEAS:HUM?\n"
        return_value = 100.0
        return_value_str = str(return_value) + "\n"
        self._transport_mock.read_until.return_value = return_value_str.encode()
        with open_close(self.instr):
            h2o = self.instr.get_humidity()

        self.assertEqual(h2o, return_value)
        self._transport_mock.write.assert_called_with(expected)

    def test_get_humidity_excepts(self):
        """Test that get_humidity call excepts on unexpected response"""
        cmd = "MEAS:HUM?"
        err = "sata"
        expected_error = "Unexpected response to {!r}, got {!r}".format(cmd, err)
        return_value_str = err + "\n"
        self._transport_mock.read_until.return_value = return_value_str.encode()
        with open_close(self.instr), self.assertRaises(qmi.core.exceptions.QMI_InstrumentException) as err:
            self.instr.get_humidity()

        self.assertEqual(str(err.exception), expected_error)


if __name__ == '__main__':
    unittest.main()



"""
Parallax Inc., USB multicore microcontroller.

The qmi.instruments.parallax package provides support for:
- USB Propeller project board
"""

from qmi.instruments.parallax.usb_propeller import Parallax_UsbPropeller



"""Unit test for Montana Cryostation S50 driver."""

import json
import unittest
from unittest.mock import MagicMock, patch

import qmi
from qmi.core.exceptions import QMI_InstrumentException

from qmi.instruments.montana import Montana_CryostationS50
from qmi.instruments.montana.cryostation_s50 import Montana_CryostationS50_System_Goal, Montana_CryostationS50_System_State


class TestCryostation50(unittest.TestCase):

    def setUp(self):
        qmi.start("montana-s50-test-context")
        self.ip_addres = "192.168.1.1"
        self.instr: Montana_CryostationS50 = qmi.make_instrument(
            "montana_s50_test_instr", Montana_CryostationS50, self.ip_addres)
        self.instr.open()
        self.controller_properties_url = f"http://{self.ip_addres}:47101/v1/controller/properties"
        self.controller_methods_url = f"http://{self.ip_addres}:47101/v1/controller/methods"
        self.cryocooler_properties_url = f"http://{self.ip_addres}:47101/v1/cooler/cryocooler/properties"
        self.sample_chamber_properties_url = f"http://{self.ip_addres}:47101/v1/sampleChamber/temperatureControllers/platform/properties"
        self.sample_chamber_thermometer_properties_url = f"http://{self.ip_addres}:47101/v1/sampleChamber/temperatureControllers/platform/thermometer/properties"

    def tearDown(self):
        self.instr.close()
        qmi.stop()

    def test_open_close(self):
        self.instr.close()
        self.assertFalse(self.instr.is_open())
        self.instr.open()
        self.assertTrue(self.instr.is_open())

    @patch('urllib.request.urlopen')
    def test_get_system_goal(self, urlopen_mock: MagicMock):
        expected_system_goal = "Vent"
        urlopen_mock.return_value.__enter__.return_value.status = 200
        urlopen_mock.return_value.__enter__.return_value.read.return_value = bytes(
            f'{{"systemGoal": "{expected_system_goal}"}}', 'utf-8')

        sys_goal = self.instr.get_system_goal()
        urlopen_mock.assert_called_once()
        req = urlopen_mock.call_args.args[0]
        self.assertEqual(req.full_url, f"{self.controller_properties_url}/systemGoal")
        self.assertEqual(req.method, 'GET')
        self.assertIsNone(req.data)
        self.assertEqual(sys_goal, Montana_CryostationS50_System_Goal.VENT)

    @patch('urllib.request.urlopen')
    def test_get_system_state(self, urlopen_mock: MagicMock):
        expected_system_state = "Ready"
        urlopen_mock.return_value.__enter__.return_value.status = 200
        urlopen_mock.return_value.__enter__.return_value.read.return_value = bytes(
            f'{{"systemState": "{expected_system_state}"}}', 'utf-8')

        sys_state = self.instr.get_system_state()
        urlopen_mock.assert_called_once()
        req = urlopen_mock.call_args.args[0]
        self.assertEqual(req.full_url, f"{self.controller_properties_url}/systemState")
        self.assertEqual(req.method, 'GET')
        self.assertIsNone(req.data)
        self.assertEqual(sys_state, Montana_CryostationS50_System_State.READY)

    @patch('urllib.request.urlopen')
    def test_cooldown_allowed(self, urlopen_mock: MagicMock):
        expected_cooldown_allowed = "true"
        urlopen_mock.return_value.__enter__.return_value.status = 200
        urlopen_mock.return_value.__enter__.return_value.read.return_value = bytes(
            f'{{"canCooldown": "{expected_cooldown_allowed}"}}', 'utf-8')

        cooldown_allowed = self.instr.cooldown_allowed()
        urlopen_mock.assert_called_once()
        req = urlopen_mock.call_args.args[0]
        self.assertEqual(req.full_url, f"{self.controller_properties_url}/canCooldown")
        self.assertEqual(req.method, 'GET')
        self.assertIsNone(req.data)
        self.assertTrue(cooldown_allowed)

    @patch('urllib.request.urlopen')
    def test_cooldown_allowed_with_http_error(self, urlopen_mock: MagicMock):
        urlopen_mock.return_value.__enter__.return_value.status = 400
        with self.assertRaises(QMI_InstrumentException):
            self.instr.cooldown_allowed()
        req = urlopen_mock.call_args.args[0]
        self.assertEqual(req.full_url, f"{self.controller_properties_url}/canCooldown")
        self.assertEqual(req.method, 'GET')
        self.assertIsNone(req.data)

    @patch('urllib.request.urlopen')
    def test_warmup_allowed(self, urlopen_mock: MagicMock):
        expected_warmup_allowed = "false"
        urlopen_mock.return_value.__enter__.return_value.status = 200
        urlopen_mock.return_value.__enter__.return_value.read.return_value = bytes(
            f'{{"canWarmup": {expected_warmup_allowed}}}', 'utf-8')

        warmup_allowed = self.instr.warmup_allowed()
        urlopen_mock.assert_called_once()
        req = urlopen_mock.call_args.args[0]
        self.assertEqual(req.full_url, f"{self.controller_properties_url}/canWarmup")
        self.assertEqual(req.method, 'GET')
        self.assertIsNone(req.data)
        self.assertFalse(warmup_allowed)

    @patch('urllib.request.urlopen')
    def test_set_platform_target_temperature(self, urlopen_mock: MagicMock):
        urlopen_mock.return_value.__enter__.return_value.status = 200

        self.instr.set_platform_target_temperature(4.9)
        urlopen_mock.assert_called_once()
        req = urlopen_mock.call_args.args[0]
        self.assertEqual(req.full_url, f"{self.controller_properties_url}/platformTargetTemperature")
        self.assertEqual(req.method, 'PUT')
        self.assertEqual(req.data, b'{\n    "platformTargetTemperature": 4.9\n}')

    @patch('urllib.request.urlopen')
    def test_set_platform_target_temperature_with_http_error(self, urlopen_mock: MagicMock):
        urlopen_mock.return_value.__enter__.return_value.status = 400

        with self.assertRaises(QMI_InstrumentException):
            self.instr.set_platform_target_temperature(4.9)
        urlopen_mock.assert_called_once()
        req = urlopen_mock.call_args.args[0]
        self.assertEqual(req.full_url, f"{self.controller_properties_url}/platformTargetTemperature")
        self.assertEqual(req.method, 'PUT')
        self.assertEqual(req.data, b'{\n    "platformTargetTemperature": 4.9\n}')

    @patch('urllib.request.urlopen')
    def test_get_platform_target_temperature(self, urlopen_mock: MagicMock):
        expected_target_temp = 4.9

        urlopen_mock.return_value.__enter__.return_value.status = 200
        urlopen_mock.return_value.__enter__.return_value.read.return_value = bytes(
            f'{{"platformTargetTemperature": {expected_target_temp}}}', 'utf-8')

        self.instr.get_platform_target_temperature()
        urlopen_mock.assert_called_once()
        req = urlopen_mock.call_args.args[0]
        self.assertEqual(req.full_url, f"{self.controller_properties_url}/platformTargetTemperature")
        self.assertEqual(req.method, 'GET')
        self.assertIsNone(req.data)

    @patch('urllib.request.urlopen')
    def test_start_cooldown(self, urlopen_mock: MagicMock):
        urlopen_mock.return_value.__enter__.return_value.status = 200
        self.instr.start_cooldown()
        urlopen_mock.assert_called_once()
        req = urlopen_mock.call_args.args[0]
        self.assertEqual(req.full_url, f"{self.controller_methods_url}/cooldown()")
        self.assertEqual(req.method, 'POST')
        self.assertIsNone(req.data)

    @patch('urllib.request.urlopen')
    def test_start_warmup(self, urlopen_mock: MagicMock):
        urlopen_mock.return_value.__enter__.return_value.status = 200
        self.instr.start_warmup()
        urlopen_mock.assert_called_once()
        req = urlopen_mock.call_args.args[0]
        self.assertEqual(req.full_url, f"{self.controller_methods_url}/warmup()")
        self.assertEqual(req.method, 'POST')
        self.assertIsNone(req.data)

    @patch('urllib.request.urlopen')
    def test_enable_platform_bakeout(self, urlopen_mock: MagicMock):
        urlopen_mock.return_value.__enter__.return_value.status = 200
        self.instr.enable_platform_bakeout()
        urlopen_mock.assert_called_once()
        req = urlopen_mock.call_args.args[0]
        self.assertEqual(req.full_url, f"{self.controller_properties_url}/platformBakeoutEnabled")
        self.assertEqual(req.method, 'PUT')
        self.assertEqual(req.data, b'{\n    "platformBakeoutEnabled": true\n}')

    @patch('urllib.request.urlopen')
    def test_enable_platform_bakeout_with_http_error(self, urlopen_mock: MagicMock):
        urlopen_mock.return_value.__enter__.return_value.status = 400

        with self.assertRaises(QMI_InstrumentException):
            self.instr.enable_platform_bakeout()
        urlopen_mock.assert_called_once()
        req = urlopen_mock.call_args.args[0]
        self.assertEqual(req.full_url, f"{self.controller_properties_url}/platformBakeoutEnabled")
        self.assertEqual(req.method, 'PUT')
        self.assertEqual(req.data, b'{\n    "platformBakeoutEnabled": true\n}')

    @patch('urllib.request.urlopen')
    def test_disable_platform_bakeout(self, urlopen_mock: MagicMock):
        urlopen_mock.return_value.__enter__.return_value.status = 200
        self.instr.disable_platform_bakeout()
        urlopen_mock.assert_called_once()
        req = urlopen_mock.call_args.args[0]
        self.assertEqual(req.full_url, f"{self.controller_properties_url}/platformBakeoutEnabled")
        self.assertEqual(req.method, 'PUT')
        self.assertEqual(req.data, b'{\n    "platformBakeoutEnabled": false\n}')

    @patch('urllib.request.urlopen')
    def test_set_platform_bakeout_temperature(self, urlopen_mock: MagicMock):
        urlopen_mock.return_value.__enter__.return_value.status = 200

        self.instr.set_platform_bakeout_temperature(312)
        urlopen_mock.assert_called_once()
        req = urlopen_mock.call_args.args[0]
        self.assertEqual(req.full_url, f"{self.controller_properties_url}/platformBakeoutTemperature")
        self.assertEqual(req.method, 'PUT')
        self.assertEqual(req.data, b'{\n    "platformBakeoutTemperature": 312\n}')

    @patch('urllib.request.urlopen')
    def test_get_platform_bakeout_temperature(self, urlopen_mock: MagicMock):
        expected_bakeout_temp = 4.9

        urlopen_mock.return_value.__enter__.return_value.status = 200
        urlopen_mock.return_value.__enter__.return_value.read.return_value = bytes(
            f'{{"platformBakeoutTemperature": {expected_bakeout_temp}}}', 'utf-8')

        bakeout_temp = self.instr.get_platform_bakeout_temperature()
        urlopen_mock.assert_called_once()
        req = urlopen_mock.call_args.args[0]
        self.assertEqual(req.full_url, f"{self.controller_properties_url}/platformBakeoutTemperature")
        self.assertEqual(req.method, 'GET')
        self.assertIsNone(req.data)
        self.assertEqual(bakeout_temp, expected_bakeout_temp)

    @patch('urllib.request.urlopen')
    def test_get_platform_bakeout_temperature_limit(self, urlopen_mock: MagicMock):
        expected_bakeout_temp_limit = 351

        urlopen_mock.return_value.__enter__.return_value.status = 200
        urlopen_mock.return_value.__enter__.return_value.read.return_value = bytes(
            f'{{"platformBakeoutTemperatureLimit": {expected_bakeout_temp_limit}}}', 'utf-8')

        bakeout_temp = self.instr.get_platform_bakeout_temperature_limit()
        urlopen_mock.assert_called_once()
        req = urlopen_mock.call_args.args[0]
        self.assertEqual(req.full_url, f"{self.controller_properties_url}/platformBakeoutTemperatureLimit")
        self.assertEqual(req.method, 'GET')
        self.assertIsNone(req.data)
        self.assertEqual(bakeout_temp, expected_bakeout_temp_limit)

    @patch('urllib.request.urlopen')
    def test_get_platform_target_temperature_limit(self, urlopen_mock: MagicMock):
        expected_temp_limit = 351

        urlopen_mock.return_value.__enter__.return_value.status = 200
        urlopen_mock.return_value.__enter__.return_value.read.return_value = bytes(
            f'{{"temperatureLimit": {expected_temp_limit}}}', 'utf-8')

        bakeout_temp = self.instr.get_platform_target_temperature_limit()
        urlopen_mock.assert_called_once()
        req = urlopen_mock.call_args.args[0]
        self.assertEqual(req.full_url, f"{self.sample_chamber_properties_url}/temperatureLimit")
        self.assertEqual(req.method, 'GET')
        self.assertIsNone(req.data)
        self.assertEqual(bakeout_temp, expected_temp_limit)

    @patch('urllib.request.urlopen')
    def test_set_platform_bakeout_time(self, urlopen_mock: MagicMock):
        urlopen_mock.return_value.__enter__.return_value.status = 200

        self.instr.set_platform_bakeout_time(120)
        urlopen_mock.assert_called_once()
        req = urlopen_mock.call_args.args[0]
        self.assertEqual(req.full_url, f"{self.controller_properties_url}/platformBakeoutTime")
        self.assertEqual(req.method, 'PUT')
        self.assertEqual(req.data, b'{\n    "platformBakeoutTime": 120\n}')

    @patch('urllib.request.urlopen')
    def test_get_platform_bakeout_time(self, urlopen_mock: MagicMock):
        expected_bakeout_time = 180

        urlopen_mock.return_value.__enter__.return_value.status = 200
        urlopen_mock.return_value.__enter__.return_value.read.return_value = bytes(
            f'{{"platformBakeoutTime": {expected_bakeout_time}}}', 'utf-8')

        bakeout_time = self.instr.get_platform_bakeout_time()
        urlopen_mock.assert_called_once()
        req = urlopen_mock.call_args.args[0]
        self.assertEqual(req.full_url, f"{self.controller_properties_url}/platformBakeoutTime")
        self.assertEqual(req.method, 'GET')
        self.assertIsNone(req.data)
        self.assertEqual(bakeout_time, expected_bakeout_time)

    @patch('urllib.request.urlopen')
    def test_is_crycooler_running(self, urlopen_mock: MagicMock):
        expected_is_cryocooler_running = "false"
        urlopen_mock.return_value.__enter__.return_value.status = 200
        urlopen_mock.return_value.__enter__.return_value.read.return_value = bytes(
            f'{{"cryocoolerRunning": {expected_is_cryocooler_running}}}', 'utf-8')

        is_cryocooler_running = self.instr.is_cryocooler_running()
        urlopen_mock.assert_called_once()
        req = urlopen_mock.call_args.args[0]
        self.assertEqual(req.full_url, f"{self.cryocooler_properties_url}/cryocoolerRunning")
        self.assertEqual(req.method, 'GET')
        self.assertIsNone(req.data)
        self.assertFalse(is_cryocooler_running)

    @patch('urllib.request.urlopen')
    def test_is_crycooler_connected(self, urlopen_mock: MagicMock):
        expected_is_cryocooler_connected = "true"
        urlopen_mock.return_value.__enter__.return_value.status = 200
        urlopen_mock.return_value.__enter__.return_value.read.return_value = bytes(
            f'{{"deviceConnected": {expected_is_cryocooler_connected}}}', 'utf-8')

        is_cryocooler_connected = self.instr.is_cryocooler_connected()
        urlopen_mock.assert_called_once()
        req = urlopen_mock.call_args.args[0]
        self.assertEqual(req.full_url, f"{self.cryocooler_properties_url}/deviceConnected")
        self.assertEqual(req.method, 'GET')
        self.assertIsNone(req.data)
        self.assertTrue(is_cryocooler_connected)

    @patch('urllib.request.urlopen')
    def test_get_sample_thermometer_properties(self, urlopen_mock: MagicMock):
        expected_resistance_ok = False
        expected_temperature_ok = True
        expected_temperature_stability_ok = False
        expected_resistance = 95
        expected_temperature = 293
        expected_temperature_stability = 10
        expected_temperature_stable = True
        json_resp = {
            "sample": {
                "resistanceOK": expected_resistance_ok,
                "resistance": expected_resistance,
                "temperatureOK": expected_temperature_ok,
                "temperature": expected_temperature,
                "temperatureStabilityOK": expected_temperature_stability_ok,
                "temperatureStability": expected_temperature_stability,
                "temperatureStable": expected_temperature_stable
            }
        }

        urlopen_mock.return_value.__enter__.return_value.status = 200
        urlopen_mock.return_value.__enter__.return_value.read.return_value = bytes(json.dumps(json_resp), 'utf-8')

        sample_data = self.instr.get_sample_thermometer_properties()
        urlopen_mock.assert_called_once()
        req = urlopen_mock.call_args.args[0]
        self.assertEqual(req.full_url, f"{self.sample_chamber_thermometer_properties_url}/sample")
        self.assertEqual(req.method, 'GET')
        self.assertIsNone(req.data)
        self.assertEqual(sample_data.resistance, expected_resistance)
        self.assertFalse(sample_data.resistance_ok)
        self.assertEqual(sample_data.temperature, expected_temperature)
        self.assertTrue(sample_data.temperature_ok)
        self.assertEqual(sample_data.temperature_stability, expected_temperature_stability)
        self.assertTrue(sample_data.temperature_stable)
        self.assertFalse(sample_data.temperature_stability_ok)


if __name__ == '__main__':
    unittest.main()



"""Unit-tests for Rohde&Schwarz SGS100a."""
from typing import cast
import unittest
from unittest.mock import patch

import qmi
from qmi.core.exceptions import QMI_InstrumentException, QMI_TimeoutException
from qmi.core.transport import QMI_TcpTransport
from qmi.instruments.rohde_schwarz import RohdeSchwarz_Sgs100a


class TestSGS100A(unittest.TestCase):

    def setUp(self):
        qmi.start("TestSGS100AContext")
        # Add patches
        patcher = patch('qmi.instruments.rohde_schwarz.rs_base_signal_gen.create_transport', spec=QMI_TcpTransport)
        self._transport_mock = patcher.start().return_value
        self.addCleanup(patcher.stop)
        patcher = patch('qmi.instruments.rohde_schwarz.rs_base_signal_gen.ScpiProtocol', autospec=True)
        self._scpi_mock = patcher.start().return_value
        self.addCleanup(patcher.stop)
        # Make DUT
        self.instr: RohdeSchwarz_Sgs100a = qmi.make_instrument("SGS100a", RohdeSchwarz_Sgs100a, "")
        self.instr = cast(RohdeSchwarz_Sgs100a, self.instr)
        self.instr.open()

    def tearDown(self):
        self.instr.close()
        qmi.stop()

    def test_start_calibration(self):
        """Test start calibration."""
        self.instr.start_calibration()

        self._scpi_mock.write.assert_called_once_with(":CAL:ALL?")

    def test_check_if_calibrating(self):
        """Test that ongoing calibration inhibits interactions with the instrument."""
        # Device doesn't respond during calibration
        self._transport_mock.read_until.side_effect = QMI_TimeoutException

        self.instr.start_calibration()

        with self.assertRaises(QMI_InstrumentException):
            self.instr.get_idn()

    def test_get_set_ref_frequency(self):
        """Test get/set external reference frequency."""
        # Test get.
        value = "10MHZ"
        self._scpi_mock.ask.return_value = value

        result = self.instr.get_external_reference_frequency()

        self._scpi_mock.ask.assert_called_once_with(":ROSC:EXT:FREQ?")
        self.assertEqual(result, value)

        # Test set.
        self._scpi_mock.ask.return_value = "0,\"No error\""
        for target_value in ("10MHZ", "100MHZ", "1000MHZ", "10MHz", "100mhz"):
            self.instr.set_external_reference_frequency(target_value)

            self._scpi_mock.write.assert_called_once_with(f":ROSC:EXT:FREQ {target_value.upper()}")
            self._scpi_mock.write.reset_mock()

        # Test invalid value.
        with self.assertRaises(ValueError):
            self.instr.set_external_reference_frequency("10000MHZ")

    def test_get_set_ref_bandwidth(self):
        """Test get/set reference bandwidth."""
        # Test get.
        value = "NARR"
        self._scpi_mock.ask.return_value = value

        result = self.instr.get_external_reference_bandwidth()

        self._scpi_mock.ask.assert_called_once_with(":ROSC:EXT:SBAN?")
        self.assertEqual(result, value)

        # Test set.
        self._scpi_mock.ask.return_value = "0,\"No error\""
        for target_value in ("NARR", "WIDE"):
            self.instr.set_external_reference_bandwidth(target_value)

            self._scpi_mock.write.assert_called_once_with(f":ROSC:EXT:SBAN {target_value}")
            self._scpi_mock.write.reset_mock()

        # Test invalid value.
        with self.assertRaises(ValueError):
            self.instr.set_external_reference_bandwidth("flat")

    def test_get_set_trigger_impedance(self):
        """Test get/set trigger impedance."""
        # Test get.
        value = "G50"
        self._scpi_mock.ask.return_value = value

        result = self.instr.get_trigger_impedance()

        self._scpi_mock.ask.assert_called_once_with(":PULM:TRIG:EXT:IMP?")
        self.assertEqual(result, value)

        # Test set.
        self._scpi_mock.ask.return_value = "0,\"No error\""
        for target_value in ("G50", "G10K"):
            self.instr.set_trigger_impedance(target_value)

            self._scpi_mock.write.assert_called_once_with(f":PULM:TRIG:EXT:IMP {target_value}")
            self._scpi_mock.write.reset_mock()

        # Test invalid value.
        with self.assertRaises(ValueError):
            self.instr.set_trigger_impedance("G10M")

    def test_get_set_iq_correction_enable(self):
        """Test get/set IQ correction enable."""
        # Test get.
        value = "1"
        self._scpi_mock.ask.return_value = value

        result = self.instr.get_iq_correction_enabled()

        self._scpi_mock.ask.assert_called_once_with(":IQ:IMP:STAT?")
        self.assertEqual(result, True)

        # Test set.
        self._scpi_mock.ask.return_value = "0,\"No error\""
        for target_value in (True, False):
            self.instr.set_iq_correction_enabled(target_value)

            self._scpi_mock.write.assert_called_once_with(":IQ:IMP:STAT {}".format(
                1 if target_value else 0
            ))
            self._scpi_mock.write.reset_mock()


if __name__ == '__main__':
    unittest.main()



import os, sys
import logging
from typing import Optional

from qmi.core.exceptions import QMI_TimeoutException, QMI_TransportDescriptorException, QMI_InstrumentException,\
    QMI_RuntimeException, QMI_EndOfInputException, QMI_Exception
from qmi.core.transport import QMI_Transport

if sys.platform != "win32":
    raise QMI_RuntimeException("The GPIB transport supports only Windows platforms.")

import usb
import usb.core
from usb.backend import libusb1
try:
    backend = libusb1.get_backend(find_library=lambda x: os.getenv("LIBUSBPATH"))
    usb.core.find(backend=backend)

except usb.core.NoBackendError as exc:
    raise QMI_RuntimeException("LIBUSBPATH environment variable must be set for libusb") from exc

import pyvisa
import pyvisa.errors

# Global variable holding the logger for this module.
_logger = logging.getLogger(__name__)


class QMI_VisaGpibTransport(QMI_Transport):
    """The QMI_VisaGpibTransport is developed specifically to use with National Instruments'
    GPIB-USB-HS device under Windows. Note that the device's GPIB settings need to be correct
    for this transport to work. Also, the presumption is that default IVI_VISA library is used.

    This transport was tested with the GPIB device connected to a Tektronix AWG 5014C instrument.
    The instrument's GPIB settings were set to:
    - GPIB interface ID: 0
    - primary address: 1
    - secondary address: None
    - timeout: 30s
    - NO sending of EOI with write nor terminating read with EOS
    - Neither System Controller nor Autopolling were enabled.
    """

    def __init__(
        self,
        primary_addr: int,
        board: Optional[int] = None,
        secondary_addr: Optional[int] = None,
        connect_timeout: float = 30.0
    ) -> None:
        """Initialization of the Gpib transport.

        Parameters:
            primary_addr: The device number to initialize.
            board: Optional interface ID, "GPIBx", number?
            secondary_addr: Optional Secondary device number.
            connect_timeout: The device timeout to open resource, in seconds. Default is 30s.
        """
        _logger.debug("Opening GPIB device nr (%i)", primary_addr)
        super().__init__()
        self._primary_addr = primary_addr
        self._board = board
        self._secondary_addr = secondary_addr
        self._connect_timeout = connect_timeout
        self._device: Optional[pyvisa.ResourceManager] = None
        self._read_buffer = bytes()

    def _open_transport(self) -> None:
        visa_resource = f"GPIB{self._board}::{self._primary_addr}::" if self._board else f"GPIB::{self._primary_addr}::"
        if self._secondary_addr:
            visa_resource = visa_resource + f"{self._secondary_addr}::"

        visa_resource = visa_resource + "INSTR"

        rm = pyvisa.ResourceManager()
        try:
            self._device = rm.open_resource(
                visa_resource,
                open_timeout=int(self._connect_timeout * 1000),
                write_termination='\n',
                read_termination='\n'
            )

        except ValueError as exc:
            if "install a suitable backend" in str(exc):
                raise QMI_TransportDescriptorException("LIBUSBPATH environment variable must be set for libusb") \
                    from exc

            else:
                raise QMI_InstrumentException(exc)

    @property
    def _safe_device(self) -> pyvisa.ResourceManager:
        """ The _safe_device property should be used inside the QMI_Transport code if-and-only-if we are 100% sure that
        the _device attribute is not None.

        This aids in static typechecking, since whereas the type of _device is Optional[T], the result of this method is
        guaranteed to be of type T. It is a QMI-internal bug if this property is used in case _device is None. In that
        case, we raise an AssertionError, and we hope the users will complain to us so we can fix the bug in the
        library.

        Raises: AssertionError: in case the property is used when the underlying value of _device is None.

        Returns: The value of _device, if it is not None. """
        assert self._device is not None
        return self._device

    def __str__(self) -> str:
        if self._board is not None:
            descriptor = f"QMI_VisaGpibTransport GPIB{self._board}::{self._primary_addr}::"
        else:
            descriptor = f"QMI_VisaGpibTransport GPIB::{self._primary_addr}::"
        descriptor = descriptor + f"{self._secondary_addr}::INSTR" if self._secondary_addr else descriptor + "INSTR"
        return descriptor

    def close(self) -> None:
        _logger.debug("Closing GPIB device nr %i", self._primary_addr)
        super().close()

    def write(self, data: bytes) -> None:
        self._check_is_open()
        self._safe_device.timeout = 0.0
        self._safe_device.write_raw(data)

    def read(self, nbytes: int, timeout: Optional[float]) -> bytes:
        self._check_is_open()

        # Read a GPIB message
        ret = self._read_message(timeout)
        if len(ret) == nbytes:
            return ret

        _logger.debug("GPIB read message contained data %s" % self._read_buffer.decode())
        raise QMI_EndOfInputException(
            f"The read message did not contain expected bytes of data ({len(ret)} != {nbytes}."
        )

    def read_until(self, message_terminator: bytes, timeout: Optional[float]) -> bytes:
        """The specified message_terminator is ignored. Instead, the instrument must autonomously indicate the end
        of the message according to the GPIB protocol.

        As the message_terminator is not used, we forward simply the call to the `read_until_timeout` call.

        Parameters:
            message_terminator: This input is ignored.
            timeout: Maximum time to wait (in seconds).

        Returns:
            Received bytes.
        """
        return self.read_until_timeout(0, timeout)

    def read_until_timeout(self, nbytes: int, timeout: Optional[float]) -> bytes:
        """Read a single USBTMC message from the instrument.

        If the timeout expires before the message is received, the read is
        aborted and any data already received are discarded. In this
        case an empty bytes string is returned.

        Parameters:
            nbytes: This input is ignored.
            timeout: Maximum time to wait (in seconds).

        Returns:
            Received bytes.
        """
        self._check_is_open()
        # Read a new message from the instrument and ignore timeout.
        try:
            data = self._read_message(timeout)
        except QMI_TimeoutException:
            data = bytes()

        return data

    def discard_read(self) -> None:
        try:
            self._read_message(0.0)
        except QMI_TimeoutException:
            return  # Nothing to discard.

    def _read_message(self, timeout: Optional[float]) -> bytes:
        """Read a GPIB message. A whole message is read from the device in bytes.

        Parameters:
            timeout: Timeout for reading
        """
        if timeout is not None:
            self._safe_device.timeout = int(timeout * 1000)  # in milliseconds

        else:
            self._safe_device.timeout = timeout  # infinite

        try:
            data = self._safe_device.read_raw()
        except pyvisa.errors.VisaIOError as exc:
            if exc.error_code == pyvisa.errors.VI_ERROR_TMO:
                raise QMI_TimeoutException(str(exc)) from exc

            raise QMI_Exception("Unknown pyvisa exception") from exc

        return data



"""
Definitions of data structures for QMI configuration data.
"""

from collections import OrderedDict
from dataclasses import field
from typing import Dict, List, Optional, Type

from qmi.core.config_struct import configstruct


@configstruct
class CfgLogging:
    """Configuration of QMI logging policy.

    Attributes:
        loglevel: Log level.
        console_loglevel: Log level for logging to console.
        logfile: Name of QMI log file, relative to QMI home directory.
        loglevels: Optional {<logger>, <level>} mapping from logger name to specific log level for that logger.
        rate_limit: Maximum number of log messages per second when logging to file.
        burst_limit: Maximum number of log messages than can be logged to file in a short burst.
    """
    loglevel:           str             = "INFO"
    console_loglevel:   str             = "WARNING"
    logfile:            str             = "qmi.log"
    loglevels:          Dict[str, str]  = field(default_factory=OrderedDict)
    rate_limit:         Optional[float] = None
    burst_limit:        int             = 1


@configstruct
class CfgContext:
    """Configuration of a QMI context.

    Attributes:
        host: IP address where the context runs (required if the context accepts peer connections).
        tcp_server_port: TCP port for incoming peer connections (or None, to disable incoming connections).
        connect_to_peers: List of peer contexts to connect to.
        enabled: True to start this context via QMI process management.
        program_module: Python module to invoke as main script.
        program_args: Optional arguments passed when starting this context.
        python_path: Optional Python search path (overrides $PYTHONPATH).
        virtualenv_path: Optional path to virtual environment to activate.
    """
    host:               Optional[str]   = None
    tcp_server_port:    Optional[int]   = None
    connect_to_peers:   List[str]       = field(default_factory=list)
    enabled:            bool            = False
    program_module:     Optional[str]   = None
    program_args:       List[str]       = field(default_factory=list)
    python_path:        Optional[str]   = None
    virtualenv_path:    Optional[str]   = None


@configstruct
class CfgProcessHost:
    """Configuration of process management for a specific host.

    Attributes:
        logdir: Directory for output logs from managed processes.
        server_command: Command to run on the host for remote process management.
        ssh_host: Optional host name to pass to SSH to access this host (default: host address).
        ssh_user: Optional user name to pass to SSH to access this host (default: current user).
    """
    server_command:     Optional[str]   = None
    ssh_host:           Optional[str]   = None
    ssh_user:           Optional[str]   = None


@configstruct
class CfgProcessManagement:
    """Configuration of process management.

    Attributes:
        output_dir: Directory for output logs from managed processes.
        hosts: Mapping from host address to process management configuration for that host.
    """
    output_dir:         Optional[str]   = None
    hosts:              Dict[str, CfgProcessHost] = field(default_factory=OrderedDict)


@configstruct
class CfgQmi:
    """Top-level QMI configuration structure.

    Attributes:
        config_file:    Absolute path name of the configuration file.
        workgroup:      Name of the QMI workgroup.
        qmi_home:       QMI home directory (or None to derive the home directory from environment settings).
        log_dir:        Directory to write various log files (or None to use the QMI home directory).
        datastore:      Location of the DataStore repository.
        logging:        Logging configuration.
        contexts:       Mapping from context name to configuration for that context.
        process_management: Process management configuration.
    """
    config_file:        Optional[str]   = None
    workgroup:          str             = "default"
    qmi_home:           Optional[str]   = None
    log_dir:            Optional[str]   = None
    datastore:          Optional[str]   = None
    logging:            CfgLogging      = field(default_factory=CfgLogging)
    contexts:           Dict[str, CfgContext] = field(default_factory=OrderedDict)
    process_management: CfgProcessManagement = field(default_factory=CfgProcessManagement)



#! /usr/bin/env python3

"""Unit test for T2 mode event processing in MultiHarp driver."""

import ctypes
import threading
import unittest
from unittest.mock import patch, PropertyMock

import numpy as np

from qmi.core.exceptions import QMI_RuntimeException
from qmi.core.pubsub import QMI_SignalReceiver
from qmi.instruments.picoquant.support._mhlib_function_signatures import _mhlib_function_signatures
from qmi.instruments.picoquant import EventDataType, EventFilterMode
from qmi.instruments.picoquant import PicoQuant_MultiHarp150

from tests.patcher import PatcherQmiContext as QMI_Context


def gen_random_events(num_events, event_rate, sync_rate, channels, seed=None):
    rng = np.random.default_rng(seed)

    mean_event_interval = 12.5e9 / event_rate
    delta_times = rng.exponential(scale=mean_event_interval, size=num_events)
    delta_times += 1.0
    delta_times = delta_times.astype(np.uint64)
    event_timestamps = np.cumsum(delta_times)

    mean_sync_interval = 12.5e9 / sync_rate
    sync_timestamps = mean_sync_interval * np.arange(1, num_events + 1)
    sync_timestamps += rng.normal(loc=0.0, scale=12.5, size=num_events)
    sync_timestamps = sync_timestamps.astype(np.uint64)

    events = np.empty(2 * num_events, dtype=EventDataType)
    events[:num_events]["type"] = rng.choice(channels, size=num_events)
    events[:num_events]["timestamp"] = event_timestamps
    events[num_events:]["type"] = 64
    events[num_events:]["timestamp"] = sync_timestamps
    events.sort(order="timestamp")

    return events[:num_events]


def events_to_fifo(events):
    """Convert array of events to the corresponding MultiHarp T2 FIFO data words."""

    # Convert events to array (if necessary).
    evt_array = np.array(events, dtype=EventDataType)

    # Check monotonic timestamps.
    assert np.all(evt_array["timestamp"][1:] >= evt_array["timestamp"][:-1])

    # Convert events to FIFO words. "type" for channel numbers and "timestamp" for time tags
    fifo_words = ((evt_array["type"].astype(np.uint32) << 25)
                  | (evt_array["timestamp"].astype(np.uint32) & 0x1ffffff))

    # Insert overflow records at positions where the MSB part of the timestamp changes.
    tmp_timestamps = np.insert(evt_array["timestamp"], 0, 0)
    (overflow_idx,) = np.where((tmp_timestamps[:-1] >> 25) != (tmp_timestamps[1:] >> 25))
    overflow_counts = (tmp_timestamps[overflow_idx + 1] >> 25) - (tmp_timestamps[overflow_idx] >> 25)
    # have to set special bit (31 from this side) for overflow
    overflow_words = overflow_counts.astype(np.uint32) | ((1 << 31) | (63 << 25))
    fifo_words = np.insert(fifo_words, overflow_idx, overflow_words)

    return fifo_words


def make_patched_read_fifo(fifo_words, done_event):
    """Return a function to be used in place of the actual readFifo() method in the MultiHarp API."""

    fifo_pos = [0]

    def patched_read_fifo(_, fifo_data, nactual):
        n_fifo_words = len(fifo_words)
        pos = fifo_pos[0]
        npos = min(n_fifo_words, pos + 10240)  # read 10240 words = 655360 bits (1.6x < 1048576). Resets timestamp.
        fifo_pos[0] = npos
        for i, val in enumerate(fifo_words[pos:npos]):
            fifo_data[i] = val
        nactual.value = npos - pos
        if (done_event is not None) and (pos == n_fifo_words):
            done_event.set()
        return 0

    return patched_read_fifo


class TestMultiHarpEventFilter(unittest.TestCase):

    # Reduced limit on maximum pending events to make it easier to test.
    PATCHED_MAX_PENDING_EVENTS = 100000
    PATCHED_MAX_EVENTS_PER_CALL = 500000  # Should be less than TTREADMAX

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Generate random event sets to be used in test cases.
        self.events_in_slow = gen_random_events(
            num_events=100000,
            event_rate=0.1e6,
            sync_rate=100.0e3,
            channels=[0, 1],
            seed=2345)

        self.events_in_fast = gen_random_events(
            num_events=500000,
            event_rate=2.0e6,
            sync_rate=100.0e3,
            channels=[0, 1],
            seed=2345)

        self.events_in_4chan = gen_random_events(
            num_events=500000,
            event_rate=1.0e6,
            sync_rate=100.0e3,
            channels=[0, 1, 2, 3],
            seed=1234)

    def setUp(self) -> None:
        patcher = patch('qmi.instruments.picoquant.PicoQuant_MultiHarp150._lib', new_callable=PropertyMock)
        self._library_mock = patcher.start().return_value
        self.addCleanup(patcher.stop)

        function_names, _, _ = zip(*_mhlib_function_signatures)
        stripped_names = [name.split('_')[1] for name in function_names]
        self._library_mock.mock_add_spec(stripped_names, spec_set=True)

        # Start QMI context.
        self._ctx = QMI_Context("multiharp_eventfilter_test")

        # Create instrument instance. Limit MAX calls to avoid problems.
        PicoQuant_MultiHarp150.MAX_EVENTS_PER_CALL = self.PATCHED_MAX_EVENTS_PER_CALL
        if self._testMethodName == "test_pending_events_limit":
            self._multiharp: PicoQuant_MultiHarp150 = PicoQuant_MultiHarp150(
                self._ctx, "multiharp", "1111111", max_pending_events=self.PATCHED_MAX_PENDING_EVENTS
            )
        else:
            self._multiharp: PicoQuant_MultiHarp150 = PicoQuant_MultiHarp150(self._ctx, "multiharp", "1111111")

        self._library_mock.GetLibraryVersion.return_value = 0
        self._library_mock.OpenDevice.return_value = 0

        string_buffer = ctypes.create_string_buffer(b'1111111')
        with patch('sys.platform', 'linux1'), patch('ctypes.create_string_buffer', return_value=string_buffer):
            self._multiharp.open()

        self._multiharp.initialize("T2", "EXTERNAL_10MHZ")

    def tearDown(self):
        # Close instrument instance.
        self._multiharp.close()
        self._ctx = None

    def test_simple(self):

        # Events to use during this test.
        events_in = [(64, 100),
                     (0, 101),
                     (1, 102),
                     (64, 1000000000),
                     (0,  1000000001),
                     (1,  1000000002),
                     (0,  1010000001),
                     (1,  1010000002),
                     (64, 1020000000)]

        # Patch the readFifo() function.
        fifo_words_in = events_to_fifo(events_in)
        done_event = threading.Event()
        self._library_mock.ReadFiFo.side_effect = make_patched_read_fifo(fifo_words_in, done_event)

        # Start the measurement.
        self._library_mock.StartMeas.return_value = 0
        self._multiharp.start_measurement(1000)

        # Wait to make sure all events are processed.
        done_event.wait()

        # Stop the measurement.
        self._multiharp.stop_measurement()

        # Fetch events.
        events = self._multiharp.get_events()

        # Check events.
        self.assertIsInstance(events, np.ndarray)
        self.assertEqual(events.dtype, EventDataType)
        self.assertTrue(np.all(events == np.array(events_in, dtype=EventDataType)))

    def test_channel_filter(self):

        # Generate a random set of events.
        events_in = self.events_in_4chan

        # Patch the readFifo() function.
        fifo_words_in = events_to_fifo(events_in)
        done_event = threading.Event()
        self._library_mock.ReadFiFo.side_effect = make_patched_read_fifo(fifo_words_in, done_event)

        # Configure event filter: block events form channels 1 and 2.
        self._multiharp.set_event_filter(channel_filter={1: EventFilterMode.NO_EVENTS,
                                                         2: EventFilterMode.NO_EVENTS})

        # Start the measurement.
        self._multiharp.start_measurement(1000)

        # Wait to make sure all events are processed.
        done_event.wait()

        # Stop the measurement.
        self._multiharp.stop_measurement()

        # Fetch events.
        events = self._multiharp.get_events()

        # Expect only events on SYNC and channels 0 and 1.
        events_expected = events_in[(events_in["type"] == 64)
                                    | (events_in["type"] == 0)
                                    | (events_in["type"] == 3)]

        # Check events.
        self.assertTrue(np.all(events == events_expected))

    def test_aperture_filter_events(self):

        # Generate a random set of events.
        events_in = self.events_in_fast

        # Patch the readFifo() function.
        fifo_words_in = events_to_fifo(events_in)
        done_event = threading.Event()
        self._library_mock.ReadFiFo.side_effect = make_patched_read_fifo(fifo_words_in, done_event)

        # Configure event filter: setup aperture filtering for channels 0 and 1.
        delta_min = 25000
        delta_max = 75000
        self._multiharp.set_event_filter(channel_filter={0: EventFilterMode.APERTURE,
                                                         1: EventFilterMode.APERTURE},
                                         sync_aperture=(delta_min, delta_max))

        # Start the measurement.
        self._multiharp.start_measurement(1000)

        # Wait to make sure all events are processed.
        done_event.wait()

        # Stop the measurement.
        self._multiharp.stop_measurement()

        # Fetch events.
        events = self._multiharp.get_events()

        # Expect only events on SYNC and events that fall in the aperture.
        events_expected = []
        last_sync = -1
        for (evt_type, evt_timestamp) in events_in:
            if evt_type == 64:
                last_sync = evt_timestamp
                events_expected.append((evt_type, evt_timestamp))
            else:
                if (last_sync >= 0) and (delta_min <= evt_timestamp - last_sync <= delta_max):
                    events_expected.append((evt_type, evt_timestamp))

        # Check events.
        self.assertTrue(np.all(events == np.array(events_expected, dtype=EventDataType)))

    def test_aperture_filter_sync(self):

        # Generate a random set of events.
        events_in = self.events_in_slow

        # Patch the readFifo() function.
        fifo_words_in = events_to_fifo(events_in)
        done_event = threading.Event()
        self._library_mock.ReadFiFo.side_effect = make_patched_read_fifo(fifo_words_in, done_event)

        # Configure event filter: setup aperture filtering for channels 0 and 1 and SYNC.
        delta_min = 25000
        delta_max = 75000
        self._multiharp.set_event_filter(channel_filter={0: EventFilterMode.APERTURE,
                                                         1: EventFilterMode.APERTURE,
                                                         64: EventFilterMode.APERTURE},
                                         sync_aperture=(delta_min, delta_max))

        # Start the measurement.
        self._multiharp.start_measurement(1000)

        # Wait to make sure all events are processed.
        done_event.wait()

        # Stop the measurement.
        self._multiharp.stop_measurement()

        # Fetch events.
        events = self._multiharp.get_events()

        # Expect only events on SYNC and events that fall in the aperture.
        events_expected = []
        last_sync = -1
        last_is_sync = False
        for (evt_type, evt_timestamp) in events_in:
            if evt_type == 64:
                if last_is_sync:
                    events_expected.pop()
                last_sync = evt_timestamp
                last_is_sync = True
                events_expected.append((evt_type, evt_timestamp))
            else:
                if (last_sync >= 0) and (delta_min <= evt_timestamp - last_sync <= delta_max):
                    events_expected.append((evt_type, evt_timestamp))
                    last_is_sync = False
        if last_is_sync:
            events_expected.pop()

        # Check events.
        self.assertTrue(np.all(events == np.array(events_expected, dtype=EventDataType)))

    def test_get_events_limit(self):
        # Test that the `get_events()` function returns at most MAX_EVENTS_PER_CALL events.

        # Generate random events.
        events_per_call = PicoQuant_MultiHarp150.MAX_EVENTS_PER_CALL
        num_events = 2 * events_per_call + 10101
        events_in = np.empty(num_events, dtype=EventDataType)
        events_in["type"] = 0
        events_in["timestamp"] = np.cumsum(np.random.randint(1, 200, len(events_in)))

        # Patch the readFifo() function.
        # The events will be passed to the driver in smaller series.
        fifo_words_in = events_to_fifo(events_in)
        done_event = threading.Event()
        self._library_mock.ReadFiFo.side_effect = make_patched_read_fifo(fifo_words_in, done_event)

        # Start the measurement.
        self._multiharp.start_measurement(1000)

        # Wait to make sure all events are processed.
        done_event.wait()

        # Stop the measurement.
        self._multiharp.stop_measurement()

        # Fetch events.
        # It should return only the first MAX_EVENTS_PER_CALL events.
        events = self._multiharp.get_events()

        # Check events.
        self.assertIsInstance(events, np.ndarray)
        self.assertEqual(events.dtype, EventDataType)
        self.assertEqual(len(events), events_per_call)
        self.assertTrue(np.all(events == events_in[:events_per_call]))

        # Now get the next MAX_EVENTS_PER_CALL events.
        (_ts, events) = self._multiharp.get_timestamped_events()
        self.assertEqual(len(events), events_per_call)
        self.assertTrue(np.all(events == events_in[events_per_call:2*events_per_call]))

        # Get the final batch of events.
        events = self._multiharp.get_events()
        self.assertEqual(len(events), 10101)
        self.assertTrue(np.all(events == events_in[2*events_per_call:]))

        # Check no further events.
        events = self._multiharp.get_events()
        self.assertIsInstance(events, np.ndarray)
        self.assertEqual(events.dtype, EventDataType)
        self.assertEqual(len(events), 0)

    def test_pending_events_limit(self):
        # Test that the driver correctly handles overflow of the pending event buffer.

        # Generate a bunch of events that just barely fit into the pending event buffer.
        events_in = np.empty(self.PATCHED_MAX_PENDING_EVENTS - 1, dtype=EventDataType)
        events_in["type"] = 0
        events_in["timestamp"] = np.cumsum(np.random.randint(1, 1000, len(events_in)))

        # Patch the readFifo() function.
        # This fill feed the first part of the events to the driver (not enough to overflow the buffer).
        fifo_words_in = events_to_fifo(events_in)
        done_event = threading.Event()
        self._library_mock.ReadFiFo.side_effect = make_patched_read_fifo(fifo_words_in, done_event)

        # Start the measurement.
        self._multiharp.start_measurement(1000)

        # Wait to make sure all queued events are processed.
        done_event.wait()

        # Fetch the events queued so far.
        events = self._multiharp.get_events()

        # Check events.
        self.assertEqual(len(events), len(events_in))
        self.assertTrue(np.all(events == events_in))

        # Generate more events than can fit into the pending event buffer.
        events_in = np.empty(self.PATCHED_MAX_PENDING_EVENTS + 1, dtype=EventDataType)
        events_in["type"] = 0
        events_in["timestamp"] = np.cumsum(np.random.randint(1, 1000, len(events_in)))

        # Repatch the readFifo() function to feed the new events into the running measurement.
        # This should/will cause the pending event buffer to overflow.
        fifo_words_in = events_to_fifo(events_in)
        done_event = threading.Event()
        self._library_mock.ReadFiFo.side_effect = make_patched_read_fifo(fifo_words_in, done_event)
        done_event.wait()

        # Try to fetch events. This should raise an exception as a result of the overflow.
        with self.assertRaises(QMI_RuntimeException):
            _ = self._multiharp.get_events()

        # Stop the measurement.
        self._multiharp.stop_measurement()

        # Generate new events to check that a subsequent measurement runs cleanly after overflow.
        events_in = np.empty(self.PATCHED_MAX_PENDING_EVENTS - 1, dtype=EventDataType)
        events_in["type"] = 0
        events_in["timestamp"] = np.cumsum(np.random.randint(1, 1000, len(events_in)))

        fifo_words_in = events_to_fifo(events_in)
        done_event = threading.Event()
        self._library_mock.ReadFiFo.side_effect = make_patched_read_fifo(fifo_words_in, done_event)

        # Start the measurement.
        self._multiharp.start_measurement(1000)

        # Wait to make sure all queued events are processed.
        done_event.wait()

        # Fetch the events queued so far.
        events = self._multiharp.get_events()

        # Check events.
        self.assertEqual(len(events), len(events_in))
        self.assertTrue(np.all(events == events_in))

        # Stop the measurement.
        self._multiharp.stop_measurement()

        # Check no further events.
        events = self._multiharp.get_events()
        self.assertEqual(len(events), 0)


class TestMultiHarpRealtime(unittest.TestCase):
    """Test class to test real-time functionalities of event processing. This requires a running QMI context.
    """
    # Reduced limit on maximum pending events to make it easier to test.
    PATCHED_MAX_PENDING_EVENTS = 100000
    PATCHED_MAX_EVENTS_PER_CALL = 500000  # Should be less than TTREADMAX

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Generate random event sets to be used in test cases.
        self.events_in_fast = gen_random_events(
            num_events=500000,
            event_rate=2.0e6,
            sync_rate=100.0e3,
            channels=[0, 1],
            seed=2345)

    def setUp(self) -> None:
        patcher = patch('qmi.instruments.picoquant.PicoQuant_MultiHarp150._lib', new_callable=PropertyMock)
        self._library_mock = patcher.start().return_value
        self.addCleanup(patcher.stop)

        function_names, _, _ = zip(*_mhlib_function_signatures)
        stripped_names = [name.split('_')[1] for name in function_names]
        self._library_mock.mock_add_spec(stripped_names, spec_set=True)

        # Start QMI context.
        self._ctx = QMI_Context("multiharp_realtime_test")
        self._ctx.start()

        # Create instrument instance. Limit MAX calls to avoid problems.
        PicoQuant_MultiHarp150.MAX_EVENTS_PER_CALL = self.PATCHED_MAX_EVENTS_PER_CALL
        if self._testMethodName == "test_pending_events_limit":
            self._multiharp = self._ctx.make_instrument(
                "multiharp", PicoQuant_MultiHarp150, "1111111", max_pending_events=self.PATCHED_MAX_PENDING_EVENTS
            )
        else:
            self._multiharp = self._ctx.make_instrument("multiharp", PicoQuant_MultiHarp150, "1111111")

        self._library_mock.GetLibraryVersion.return_value = 0
        self._library_mock.OpenDevice.return_value = 0

        string_buffer = ctypes.create_string_buffer(b'1111111')
        with patch('sys.platform', 'linux1'), patch('ctypes.create_string_buffer', return_value=string_buffer):
            self._multiharp.open()

        self._multiharp.initialize("T2", "EXTERNAL_10MHZ")

    def tearDown(self):
        # Close instrument instance.
        self._multiharp.close()
        # Stop QMI context.
        self._ctx.stop()
        self._ctx = None

    def test_realtime_histogram(self):

        # Generate a random set of events.
        events_in = self.events_in_fast

        # Patch the readFifo() function.
        fifo_words_in = events_to_fifo(events_in)
        done_event = threading.Event()
        self._library_mock.ReadFiFo.side_effect = make_patched_read_fifo(fifo_words_in, done_event)

        # Subscribe to real-time histograms.
        recv = QMI_SignalReceiver()
        self._multiharp.sig_histogram.subscribe(recv)

        # Configure event filter to reject events on both channels.
        # This should not affect the histograms.
        self._multiharp.set_event_filter(channel_filter={0: EventFilterMode.NO_EVENTS,
                                                         1: EventFilterMode.NO_EVENTS})

        # Configure real-time histograms.
        bin_resolution = 25
        histogram_bins = 5000
        histogram_num_sync = 100
        self._multiharp.set_realtime_histogram(
            channels=[0, 1],
            bin_resolution=bin_resolution,
            num_bins=histogram_bins,
            num_sync=histogram_num_sync)

        # Start the measurement.
        self._multiharp.start_measurement(1000)

        # Wait to make sure all events are processed.
        done_event.wait()

        # Stop the measurement.
        self._multiharp.stop_measurement()

        # Fetch events.
        events = self._multiharp.get_events()

        # Expect only SYNC events.
        events_expected = events_in[(events_in["type"] == 64)]
        self.assertTrue(np.all(events == events_expected))

        # Check histograms.
        sync_count = 0
        last_sync = -1
        start_timestamp = 0
        for (evt_type, evt_timestamp) in events_in:
            if evt_type == 64:
                if sync_count % histogram_num_sync == 0:
                    if sync_count > 0:
                        # Check against published histogram.
                        sig = recv.get_next_signal()
                        (hist,) = sig.args
                        self.assertEqual(hist.start_timestamp, start_timestamp)
                        self.assertEqual(hist.bin_resolution, bin_resolution)
                        self.assertEqual(hist.num_sync, histogram_num_sync)
                        self.assertEqual(hist.channels, [0, 1])
                        self.assertTrue(np.all(hist.histogram_data == expect_histogram))
                    expect_histogram = np.zeros((2, histogram_bins), dtype=np.uint32)
                    start_timestamp = evt_timestamp
                sync_count += 1
                last_sync = int(evt_timestamp)
            elif sync_count > 0:
                # Count event in expected histogram.
                bin_index = min((int(evt_timestamp) - last_sync) // bin_resolution, histogram_bins - 1)
                expect_histogram[evt_type, bin_index] += 1

        # Check that there are no further published histograms.
        self.assertFalse(recv.has_signal_ready())

        self._multiharp.sig_histogram.unsubscribe(recv)

    def test_realtime_countrate(self):

        # Generate a random set of events.
        events_in = self.events_in_fast

        # Patch the readFifo() function.
        fifo_words_in = events_to_fifo(events_in)
        done_event = threading.Event()
        self._library_mock.ReadFiFo.side_effect = make_patched_read_fifo(fifo_words_in, done_event)

        # Subscribe to real-time count rate.
        recv = QMI_SignalReceiver()
        self._multiharp.sig_countrate.subscribe(recv)

        # Block all events.
        # This should not affect the count rate reports.
        self._multiharp.set_block_events(True)

        # Configure real-time count rates.
        delta_min = 25000
        delta_max = 75000
        countrate_num_sync = 80
        self._multiharp.set_realtime_countrate(sync_aperture=(delta_min, delta_max), num_sync=countrate_num_sync)

        # Start the measurement.
        self._multiharp.start_measurement(1000)

        # Wait to make sure all events are processed.
        done_event.wait()

        # Stop the measurement.
        self._multiharp.stop_measurement()

        # Fetch events.
        events = self._multiharp.get_events()

        # Expect no events (all events blocked).
        self.assertEqual(len(events), 0)

        # Check count rates.
        sync_count = 0
        last_sync = -1
        start_timestamp = 0
        for (evt_type, evt_timestamp) in events_in:
            if evt_type == 64:
                if sync_count % countrate_num_sync == 0:
                    if sync_count > 0:
                        # Check against published count rate.
                        sig = recv.get_next_signal()
                        (counts,) = sig.args
                        self.assertEqual(counts.start_timestamp, start_timestamp)
                        self.assertEqual(counts.num_sync, countrate_num_sync)
                        self.assertTrue(np.all(counts.counts == expect_counts))
                    expect_counts = np.zeros(8, dtype=np.uint64)
                    start_timestamp = evt_timestamp
                sync_count += 1
                last_sync = int(evt_timestamp)
            elif sync_count > 0:
                # Check aperture filter.
                if delta_min <= int(evt_timestamp) - last_sync <= delta_max:
                    # Count this event.
                    expect_counts[evt_type] += 1

        # Check that there are no further published count rates.
        self.assertFalse(recv.has_signal_ready())

        self._multiharp.sig_histogram.unsubscribe(recv)


if __name__ == "__main__":
    unittest.main()



"""Parse an ADbasic program.

This module provides functions to parse an ADbasic program and its include files,
extracting information from the symbol definitions in the program.

This module can also be executed with a toplevel ADbasic filename as an argument, to get an overview
of the defined symbols.
"""

import argparse
import logging
import os.path
import re
from typing import Dict, List, NamedTuple, Optional, Tuple, Union

from qmi.core.exceptions import QMI_Exception


# Global variable holding the logger for this module.
_logger = logging.getLogger(__name__)


class SymbolInfo(NamedTuple):
    """Symbol defined in the ADbasic program source."""
    filename: str
    line_nr: int
    label: str
    value: str


class ParDesc(NamedTuple):
    """Represents a global Par_nn variable."""
    par_index: int

    def __str__(self) -> str:
        return "Par_{}".format(self.par_index)


class FParDesc(NamedTuple):
    """Represents a global FPar_nn variable."""
    fpar_index: int

    def __str__(self) -> str:
        return "FPar_{}".format(self.fpar_index)


class ArrayElemDesc(NamedTuple):
    """Represents an element inside a global Data_nnn array."""
    data_index: int
    elem_index: int

    def __str__(self) -> str:
        return "Data_{}[{}]".format(self.data_index, self.elem_index)


ParameterDesc = Union[ParDesc, FParDesc, ArrayElemDesc]


class ParameterInfo(NamedTuple):
    """Represents the collection of externally accessible parameters of an ADbasic program.

    Attributes:
        param:  Mapping from parameter name to a `ParDesc`, `FParDesc` or `ArrayElemDesc` instance.
        data:   Mapping from data name to the index of the global `Data_nnn` array.
    """
    param: Dict[str, ParameterDesc]
    data: Dict[str, int]


class ParseException(QMI_Exception):
    """Raised when parsing fails."""

    def __init__(self, filename: str, line_nr: int, message: str) -> None:
        super().__init__([filename, line_nr, message])
        self.filename = filename
        self.line_nr = line_nr
        self.message = message

    def __str__(self) -> str:
        return "{}, line {}: {}".format(self.filename, self.line_nr, self.message)


def _resolve_include_path(include_path: str, source_file: str, include_dir: str) -> Optional[str]:
    """Resolve the OS path name of an ADbasic include file.

    Only local include paths are processed.
    System-wide includes are rejected and return None.

    Parameters:
        include_path: Include path string as it occurs in the ADbasic source code.
        source_file: File name of the including source file.
        include_dir: Base directory for resolving relative include paths.

    Returns:
        Path name of the OS-level location of the include file, or None if
        this include file should not be processed.
    """

    # Separate include path into components.
    path_components = include_path.replace("\\", "/").split("/")

    # Check whether the include path consists of only a file name without a directory component.
    # This means it is a system-wide include, so ignore it.
    if len(path_components) <= 1:
        return None

    # Check whether the include path is an absolute path (starts with a "/").
    # We don't support that, so warn and ignore.
    if path_components[0] == "":
        _logger.warning("Absolute include path not supported: %r", include_path)
        return None

    # Check whether the include path uses a parent component "..", or current path component ".".
    if any(comp.startswith("..") or comp.startswith(".") for comp in path_components):
        source_path = os.path.dirname(source_file)
        parent_relative_path = os.path.join(source_path, *path_components)
        norm_relative_path = os.path.normpath(parent_relative_path)
        return norm_relative_path

    # Resolve the include path relative to the local include directory.
    return os.path.join(include_dir, *path_components)


def _parse_single_adbasic_file(filename: str) -> Tuple[List[SymbolInfo], List[str]]:
    """Parse a single ADBasic source file and return a tuple of defined symbols and included files.

    Parameters:
        filename: File name of ADBasic source to parse.

    Returns:
        Tuple `(defined_symbols, include_paths)`
        where `defined_symbols` is a list of defined symbols and `include_paths` is a list of included files.
    """

    _logger.debug("Parsing %s", filename)

    with open(filename, "r") as fi:
        source = fi.read()

    # Define a regexp for lines that look like:
    #
    #   #Define SymbolName SymbolValue ' optional comment.
    #
    # This regexp is slightly complex because we have to handle an optional comment at the end of the line.

    re_define = re.compile(r"^\s*#Define\s+(\S+)\s+([^\s']+)\s*(?:'.*)?$", re.ASCII | re.IGNORECASE)

    re_include = re.compile(r"^\s*#Include\s+([^\s']+)\s*(?:'.*)?$", re.ASCII | re.IGNORECASE)

    defined_symbols = []
    include_paths = []

    for (line_nr, line_text) in enumerate(source.splitlines()):

        match = re_define.match(line_text)
        if match is not None:
            (symbol, value) = match.groups()
            symbol_info = SymbolInfo(filename, line_nr + 1, symbol, value)
            defined_symbols.append(symbol_info)

        match = re_include.match(line_text)
        if match is not None:
            include_path = match.group(1)
            include_paths.append(include_path)

    return (defined_symbols, include_paths)


def parse_adbasic_program(filename: str, include_dir: str) -> List[SymbolInfo]:
    """Parse an ADbasic program (and, recursively, its include files) to find #Define lines.

    Parameters:
        filename: File name of ADBasic source to parse.
        include_dir: Base directory for resolving relative include paths.
                     May be empty to use the current working directory.

    Returns:
        List of SymbolInfo objects to describe defined symbols.
    """

    files_remaining = [filename]

    all_defined_symbols = []

    while files_remaining:
        sourcefile = files_remaining.pop(0)
        (defined_symbols, include_paths) = _parse_single_adbasic_file(sourcefile)
        all_defined_symbols.extend(defined_symbols)

        # Add included files to parse queue.
        for include_path in include_paths:
            resolved_include_path = _resolve_include_path(include_path, sourcefile, include_dir)
            if resolved_include_path:
                files_remaining.append(resolved_include_path)

    return all_defined_symbols


def _extract_data_defines(symbols: List[SymbolInfo]) -> Dict[str, int]:
    """Extract data array information from symbol definitions."""

    # Maintain a mapping from uppercase parameter name to the actual case of the name.
    # This is used to check for duplicate definitions because identifiers are case-insensitive in ADbasic.
    name_case_map: Dict[str, str] = {}

    data_info: Dict[str, int] = {}
    ref_to_name: Dict[int, str] = {}

    for symbol in symbols:

        # Only process symbol names that start with "DATA_".
        if not symbol.label.upper().startswith("DATA_"):
            continue

        # Extract data name.
        (_prefix, data_name) = symbol.label.split("_", maxsplit=1)

        # Extract index of the global data array.
        match = re.match(r'^Data_([0-9]+)$', symbol.value, re.IGNORECASE)
        if match:
            data_index = int(match.group(1))
        else:
            _logger.warning("Unrecognized symbol definition format for %r", symbol.label)
            continue

        # Check for duplicate definition of this symbol with different case.
        prev_name = name_case_map.get(data_name.upper())
        if (prev_name is not None) and (prev_name != data_name):
            raise ParseException(
                filename=symbol.filename,
                line_nr=symbol.line_nr,
                message="Duplicate definition of symbol {} with different case".format(symbol.label))

        # Check for duplicate definition of this symbol with different data index.
        prev_index = data_info.get(data_name)
        if (prev_index is not None) and (prev_index != data_index):
            raise ParseException(
                filename=symbol.filename,
                line_nr=symbol.line_nr,
                message="Duplicate definition of symbol {} for different data array".format(symbol.label))

        # Check for duplicate use of the same Data_nnn array under different names.
        prev_name = ref_to_name.get(data_index)
        if (prev_name is not None) and (prev_name != data_name):
            raise ParseException(
                filename=symbol.filename,
                line_nr=symbol.line_nr,
                message="Symbol {} is a duplicate reference to {}".format(symbol.label, symbol.value))

        # Store new symbol definition.
        data_info[data_name] = data_index
        ref_to_name[data_index] = data_name
        name_case_map[data_name.upper()] = data_name

    return data_info


def _extract_par_defines(symbols: List[SymbolInfo], data_info: Dict[str, int]) -> Dict[str, ParameterDesc]:
    """Extract scalar parameter information from symbol definitions."""

    # Rebuild Data_nnn index using uppercase parameter names for case-insensitive lookup.
    data_info_upper = dict((name.upper(), value) for (name, value) in data_info.items())

    # Maintain a mapping from uppercase parameter name to the actual case of the name.
    # This is used to check for duplicate definitions because identifiers are case-insensitive in ADbasic.
    name_case_map: Dict[str, str] = {}

    param_info: Dict[str, ParameterDesc] = {}
    ref_to_name: Dict[str, str] = {}

    for symbol in symbols:

        # Only process symbol names that start with "PAR_".
        if not symbol.label.upper().startswith("PAR_"):
            continue

        # Extract parameter name.
        (_prefix, param_name) = symbol.label.split("_", maxsplit=1)

        param_desc: Optional[ParameterDesc] = None

        # Try to match parameter in global Par variable.
        if param_desc is None:
            match = re.match(r'^Par_([0-9]+)$', symbol.value, re.IGNORECASE)
            if match:
                par_index = int(match.group(1))
                param_desc = ParDesc(par_index)

        # Try to match parameter in global FPar variable.
        if param_desc is None:
            match = re.match(r'^FPar_([0-9]+)$', symbol.value, re.IGNORECASE)
            if match:
                par_index = int(match.group(1))
                param_desc = FParDesc(par_index)

        # Try to match parameter in data array element.
        if param_desc is None:
            match = re.match(r'^Data_(\S+)\s*\[\s*([0-9]+)\s*]$', symbol.value, re.IGNORECASE)
            if match:
                data_name = match.group(1)
                elem_index = int(match.group(2))
                data_index = data_info_upper.get(data_name.upper())
                if data_index is None:
                    raise ParseException(
                        filename=symbol.filename,
                        line_nr=symbol.line_nr,
                        message="Symbol {} refers to unknown array {}".format(symbol.label, data_name))
                data_index = data_info_upper[data_name.upper()]
                param_desc = ArrayElemDesc(data_index, elem_index)

        if param_desc is None:
            _logger.warning("Unrecognized symbol definition format for %r", symbol.label)
            continue

        # Check for duplicate definition of this symbol with different case.
        prev_name = name_case_map.get(param_name.upper())
        if (prev_name is not None) and (prev_name != param_name):
            raise ParseException(
                filename=symbol.filename,
                line_nr=symbol.line_nr,
                message="Duplicate definition of symbol {} with different case".format(symbol.label))

        # Check for duplicate definition of this symbol with different data index.
        prev_desc = param_info.get(param_name)
        if (prev_desc is not None) and (type(prev_desc) != type(param_desc) or prev_desc != param_desc):
            raise ParseException(
                filename=symbol.filename,
                line_nr=symbol.line_nr,
                message="Duplicate definition of symbol {} for different parameter".format(symbol.label))

        # Check for duplicate use of the same parameter under different names.
        prev_name = ref_to_name.get(repr(param_desc))
        if (prev_name is not None) and (prev_name != param_name):
            raise ParseException(
                filename=symbol.filename,
                line_nr=symbol.line_nr,
                message="Symbol {} is a duplicate reference to {}".format(symbol.label, symbol.value))

        # Store new symbol definition.
        param_info[param_name] = param_desc
        ref_to_name[repr(param_desc)] = param_name
        name_case_map[param_name.upper()] = param_name

    return param_info


def analyze_parameter_info(symbols: List[SymbolInfo]) -> ParameterInfo:
    """Analyze a set of ADbasic symbol definitions to extract parameter information.

    Only the following types of symbol definitions are processed:
     - `#Define PAR_parname Par_nn` (binds `parname` to global Par variable).
     - `#Define PAR_parname FPar_nn` (binds `parname` to global FPar variable).
     - `#Define DATA_arrayname Data_nnn` (binds `arrayname` to global Data array).
     - `#Define PAR_parname DATA_arrayname[ii]` (binds `parname` to a specific element in a global data array).

    Parameters:
        symbols: List of symbol definitions (as produced by `parse_adbasic_program`).

    Returns:
        ParameterInfo instance containing a description of all named parameters and named arrays.
    """

    data_info = _extract_data_defines(symbols)
    param_info = _extract_par_defines(symbols, data_info)
    return ParameterInfo(param_info, data_info)


def print_symbol_info(symbols: List[SymbolInfo]) -> None:
    """Print the result of parse_adbasic_program()."""

    for symbol in symbols:
        print("{}, {}: #Define {} {}".format(symbol.filename, symbol.line_nr, symbol.label, symbol.value))


def print_parameter_info(param_info: ParameterInfo) -> None:
    """Print the result of analyze_adbasic_symbols()."""

    sort_order = {
        ParDesc: 1,
        FParDesc: 2,
        ArrayElemDesc: 3 }

    par_items = list(param_info.param.items())
    par_items.sort(key=lambda kv: (sort_order.get(type(kv[1])), tuple(kv[1])))
    for (name, desc) in par_items:
        print("PAR_{:32} = {}".format(name, str(desc)))

    data_items = list(param_info.data.items())
    data_items.sort(key=lambda kv: kv[1])
    for (name, index) in data_items:
        print("DATA_{:32} = Data_{}".format(name, index))


def main():

    parser = argparse.ArgumentParser(description="Parse ADbasic program and extract #Define'd constants.")
    parser.add_argument("filename", help="toplevel ADbasic file to be analyzed")
    args = parser.parse_args()

    symbols = parse_adbasic_program(args.filename, os.path.dirname(args.filename))
    #print_symbol_info(symbols)

    param_info = analyze_parameter_info(symbols)
    print_parameter_info(param_info)


if __name__ == "__main__":
    main()
